'''B{Pygplot:  A Python Interface to Pgplot}
@author:  Chris Burns
@version: 0.92
@requires: Pgplot (U{http://www.astro.caltech.edu/~tjp/pgplot/})

You can get the source for pygplot here: 
U{http://astro.swarthmore.edu/~burns/pygplot_downloads}.

There is the beginnings of a manual here:
U{http://astro.swarthmore.edu/~burns/pygplot.pdf}.
                            
This python module provides an object-oriented interface to the Pgplot
library of routines.  Most (though not all) of the functionality of Pgplot 
is included:

 1. 1-D plots using lines, symbols and error-bars
 2. 2-D plots using color (or gray-scale) images and contours
 3. Access to all the pgplot primitives through the ppgplot extension module.

'''
import types, math, ppgplot
import sys, string
import colormaps
import textwrap as tw
import pickle
import re

def max_hack(array):
   '''A quick hack to get around the fact that numarray doesn't have a function
   max, only a member function.'''
   return array.max()

def min_hack(array):
   '''ditto'''
   return array.min()


# Test to see if we should import Numpy, Numeric or numarray
module = ppgplot.get_num()
if module == "numpy":
   import numpy as Numeric
   import numpy.oldnumeric.mlab as MLab
   nummin = MLab.min
   nummax = MLab.max
elif module == "numarray":
   import numarray as Numeric
   nummax = max_hack
   nummin = min_hack
else:
   import Numeric
   import MLab
   nummin = MLab.min
   nummax = MLab.max

# Default font to use
_default_font = 1
# Default character size
_default_fsize = 1.0
# Default line style to use
_default_linestyle = 1
# Default line width to use
_default_linewidth = 1
# Default line width for everything
_default_lw = 2
# Default symbol to plot
_default_symbol = 1
# Default label color for contour lines
_default_clab_color = 'white'
# Default label interval for labeling contour lines
_default_clab_int = 40
# Default minimum label interval for labeling contour lines
_default_clab_min = 10
# Default line colors to use
_default_color = 'white'
# Default color palette for IMAG routines
_default_palette = 'real'
# Default transfer function for IMAG routines
_default_transfer = 'linear'
# The set of colors for PGPLOT
_color_table = { \
   'black':0, 'Black':0, 'BLACK':0, 'bg':0, \
   'white':1, 'White':1, 'WHITE':1, 'fg':1, \
   'red':2, 'Red':2, 'RED':2, \
   'green':3, 'Green':3, 'GREEN':3, \
   'blue':4, 'Blue':4, 'BLUE':4, \
   'cyan':5, 'Cyan':5, 'CYAN':5, \
   'magenta':6, 'Magenta':6, 'MAGENTA':6, \
   'yellow':7, 'Yellow':7, 'YELLOW':7, \
   'orange':8, 'Orange':8, 'ORANGE':8, \
   'green2':9, 'Green2':9, 'GREEN2':9, \
   'green3':10, 'Green3':10, 'GREEN3':10, \
   'blue2':11, 'Blue2':14, 'BLUE2':11, \
   'purple':12, 'Purple':12, 'PURPLE':12, \
   'pink':13, 'Pink':13, 'PINK':13, \
   'darkgray':14, 'DarkGray':14, 'DARKGRAY':14, \
   'dark gray':14, 'Dark Gray':14, 'DARK GRAY':14, \
   'lightgray':15, 'LightGray':15, 'LIGHTGRAY':15, \
   'light gray':15, 'Light Gray':15, 'LIGHT GRAY':15 \
   }

# Get list of available devices
n_devices = ppgplot.pgqndt()
devices = []
descriptions = {}
for i in range(1,n_devices+1):
   #  for some reason, pgqdt doesn't seem to be totally reliable, so use
   #  a try structure
   try:
      (label,descr,inter) = ppgplot.pgqdt(i)
      devices.append(label)
      descriptions[label]=[descr,inter]
   except:
      pass

if '/XWIN' in devices:
   _default_device="/XWIN"
elif '/XSERVE' in devices:
   _default_device="/XSERVE"
elif '/XWINDOW' in devices:
   _default_device="/XWINDOW"
elif '/AQT' in devices:  
   _default_device="/AQT"
else:
   fallback = "/NULL"
   # look for the first interractive device we can find
   for dev in devices:
      if descriptions[dev][1] > 0:
         fallback = dev
         break
   _default_device = fallback

def max_hack(array):
   '''A quick hack to get around the fact that numarray doesn't have a function
   max, only a member function.'''
   return array.max()

def min_hack(array):
   '''ditto'''
   return array.min()

def list_devices():
   '''Print out the list of available devices.'''
   for dev in devices:
      print dev
      print descriptions[dev][0]
      if descriptions[dev][1]:
         print 'This is an interactive device'

def Linear(low,high,n):
   '''Convenience function for computing I{n} numbers linearly interpolated
      between I{low} and I{high}.  Used as the default function for the
      I{cfunc} parameter to the I{contour} function.'''
   a = Numeric.arange(n)
   a = float(low) + a*float(high - low)/n
   return a

def redraw(x,y,key,inst):
   inst.replot()

def printcoords(x, y, key, inst):
   print x,y

def pan(x, y, key, inst):
   '''Pan so that (x,y) is new center.'''
   if inst.xrange is None:
      rangex = Numeric.array([inst._xmin, inst._xmax])
   else:
      rangex = Numeric.array(inst.xrange)

   if inst.yrange is None:
      rangey = Numeric.array([inst._ymin, inst._ymax])
   else:
      rangey = Numeric.array(inst.yrange)

   dx = x - (rangex[0] + rangex[1])/2
   dy = y - (rangey[0] + rangey[1])/2
   rangex = rangex + dx
   rangey = rangey + dy
   inst.xrange = rangex.tolist()
   inst.yrange = rangey.tolist()
   #self.replot()

def zoom(x,y,key,inst):
   '''Zoom in or out by 50%.'''
   if inst.xrange is None:
      rangex = [inst._xmin, inst._xmax]
   else:
      rangex = inst.xrange

   if inst.yrange is None:
      rangey = [inst._ymin, inst._ymax]
   else:
      rangey = inst.yrange

   if key == 'z':  factor = 0.5
   else: factor = 1.5
   new_rangex = [0.5*(rangex[0]*(1+factor) + rangex[1]*(1-factor)),
                 0.5*(rangex[1]*(1+factor) + rangex[0]*(1-factor))]
   new_rangey = [0.5*(rangey[0]*(1+factor) + rangey[1]*(1-factor)),
                 0.5*(rangey[1]*(1+factor) + rangey[0]*(1-factor))]
   inst.xrange = new_rangex
   inst.yrange = new_rangey

def xrange_select(x,y,key,inst, tol=1e-6):
   '''Select an xrange.'''
   x2,y2,key = ppgplot.pgband(4,0,x,y)
   if inst.xrange is None:
      scale = abs(inst._xmax - inst._xmin)
   else:
      scale = abs(inst.xrange[1] - inst.xrange[0])
   if abs(x - x2)/scale < tol:
      inst.xrange = None
   else:
      inst.xrange = [x,x2]

def yrange_select(x,y,key,inst, tol=1.e-6):
   '''Select a yrange.'''
   x2,y2,key = ppgplot.pgband(3,0,x,y)
   if inst.yrange is None:
      scale = abs(inst._ymax - inst._ymin)
   else:
      scale = abs(inst.yrange[1] - inst.yrange[0])
   if abs(y - y2)/scale < tol:
      inst.yrange = None
   else:
      inst.yrange = [y,y2]

def box_range_select(x,y,key,inst, tol=1.e-6):
   '''Select an x and y range simultaneously.'''
   x2,y2,key = ppgplot.pgband(2,0,x,y)
   if inst.xrange is None:
      xscale = abs(inst._xmax - inst._xmin)
   else:
      xscale = abs(inst.xrange[1] - inst.xrange[0])
   if inst.yrange is None:
      yscale = abs(inst._ymax - inst._ymin)
   else:
      yscale = abs(inst.yrange[1] - inst.yrange[0])
   if abs(y - y2)/yscale < tol or abs(x - x2)/xscale < tol:
      inst.xrange = inst.yrange = None
   else:
      inst.xrange = [x,x2]
      inst.yrange = [y,y2]


class _Layout:
   ''' A class that handles the page layout of the graph.  Instead of using pgsubp,
   which only allows equal splitting of the page, we're going to make a custom version
   that allows for different sized panels.  It keeps track of the width/height of each
   cell and uses pgsvp to do the same effect as pgpage() or pgpanl().'''

   def __init__(self):
      self.ncells = 0       # the number of cells in the page layout
      self.urcs = []        # Coordinate of the upper-right corners of each panel
      self.llcs = []        # Coordinates of the lower-left corners of each panel
      self.v1s = []         #  coordinates of the viewports
      self.v2s = []
      self.u1s = []
      self.u2s = []
      self.current = -1     # Current cell number.  Initially set to -1 to emulate
                            # pgpage, which advances to next page OR to the first
                            # panel

   def add_panel(self, llc, urc):
      '''Add a panel to the layout.  Right now the user can over-lay panels... do we
         want this?  The only checking done is to make sure that the panel is 
         entirely on the device.'''
      try:
         x1 = llc[0]
         y1 = llc[1]
         x2 = urc[0]
         y2 = urc[1]
      except:
         raise TypeError, "urc and llc must be two-tuples"

      if not (0 <= x1 <= 1 and 0 <= x2 <= 1 and 0 <= y1 <= 1 and 0 <= y2 <= 1):
         raise ValueError, "All coordinates must be normalized device coordinates"

      # Make sure the ranges make sense.
      if x1 > x2:  (x1,x2) = (x2,x1)
      if y1 > y2:  (y1,y2) = (y2,y1)

      self.urcs.append([x2,y2])
      self.llcs.append([x1,y1])
      self.u1s.append(x1);  self.u2s.append(x2);  
      self.v1s.append(y1);  self.v2s.append(y2)
      self.ncells += 1

   def select(self, i):
      '''Select current panel.  Kind of like pgpanl.'''
      if i >= self.ncells:
         raise ValueError, "There are only %d cells." % (self.ncells)
      self.current = i

   def pgpage(self):
      '''Select the next panel, or do an actual pgpage if we're on the last panel.'''
      if self.current == self.ncells - 1:
         ppgplot.pgpage()
         self.current = 0
      else:
         self.current += 1

   def pgsvp(self, vx1, vx2, vy1, vy2):
      '''Set the viewport inside the panel.  Normally, the user would do this via
      pgsubp and then expect the pgsvp coordinates to be normalized inside that panel,
      so vx1,vx2,vy1 and vy2 are all normalized, but relative to the viewport.'''

      # The box coordinates:
      x0 = self.llcs[self.current][0]
      y0 = self.llcs[self.current][1]
      x1 = self.urcs[self.current][0]
      y1 = self.urcs[self.current][1]

      xb1 = vx1;  xb2 = 1-vx2;   yb1 = vy1;   yb2 = 1-vy2

      # now convert the input vx1...vy2  to these new sub-panel coordinates
      self.u1s[self.current] = x0 + xb1
      self.v1s[self.current] = y0 + yb1
      self.u2s[self.current] = x1 - xb2
      self.v2s[self.current] = y1 - yb2

      # Finally, call the real pgsvp:
      ppgplot.pgsvp(self.u1s[self.current],self.u2s[self.current],
            self.v1s[self.current],self.v2s[self.current])
      
   def pgqvp(self, i=None):
      '''drop-in replacement for pgqvp.'''
      if i is None:  i = self.current
      u1 = self.u1s[i]
      u2 = self.u2s[i]
      v1 = self.v1s[i]
      v2 = self.v2s[i]
      return(u1,u2,v1,v2)

   def add_border(self, bx1=0, bx2=0, by1=0, by2=0):
      '''Adjust all the coordinates of all the panels to add border space around the
      entire page (border sizes are measured in normalized device coordinates). 
      Successive calls should have a cumulative effect.'''
      # By what factor are we decrementing the size of plotting area?
      ydec = 1 - (by1 + by2)
      xdec = 1 - (bx1 + bx2)

      for i in range(self.ncells):
         self.urcs[i][0] = self.urcs[i][0]*xdec + bx1
         self.urcs[i][1] = self.urcs[i][1]*ydec + by1
         self.llcs[i][0] = self.llcs[i][0]*xdec + bx1
         self.llcs[i][1] = self.llcs[i][1]*ydec + by1

   def which_vp(self, vx, vy):
      '''Given a normalized device coordinate (vx,vy), return which viewport, if any,
      corresponds to this position.  If none, returns -1.  Assume that there is
      no overlap of cells.'''

      for i in range(self.ncells):
         if (self.llcs[i][0] < vx < self.urcs[i][0]) and \
               (self.llcs[i][1] < vy < self.urcs[i][1]):
            return i
      return -1

class _Object:
   '''An object in the plot.  The other objects will inherit from this base
   class.'''

   autorange = 1
   order = -1

   def __init__(self, **kw):
      if 'autorange' in kw:
         self.autorange = kw['autorange']
      if 'order' in kw:
         self.order = kw['order']

   def __getattr__(self, name):
      #print "requesting ",name
      if name == 'size':
         if self.parent.options['size'] is None:
            return self.parent.options['fsize']
         else:
            return self.parent.options['size']
      #if name == 'order':
      #   return -1
      if name in ['font','color','fsize','linewidth','linestyle','size','symbol']:
         return self.parent.options[name]
      #if name == 'autorange':
      #   return 1
      raise AttributeError, "Attribute %s not defined" % name

class _Axis(_Object):
   ''' A class used by the Plot class.  It defines an alternate axis, which
       is created using PGAXIS.  This does not alter the viewport or otherwise
       alter the plotting;  it is simply a convenient way to draw an axis.'''
   def __init__(self, parent, x1, y1, x2, y2, v1, v2, options='',
                step=0, nsub=0, dmajl=0.5, dmajr=0.5, fmin=0.5, disp=0.3,
                orient=0.0, fsize=None, font=None, linestyle=None, linewidth=None,
                color=None, **kw):
      '''The axis initialization routine.'''
      self.parent = parent
      self.x1 = x1;  self.x2 = x2;  self.y1 = y1;   self.y2 = y2
      self.v1 = v1;  self.v2 = v2
      self.options = options
      self.step = step
      self.nsub = nsub
      self.dmajl = dmajl; self.dmajr = dmajr
      self.fmin = fmin
      self.disp = disp
      self.orient = orient

      if fsize:  self.fsize = fsize  
      if font:  self.font = font 
      if linestyle:  self.linestyle = linestyle
      if linewidth:  self.linewidth = linewidth
      if color:  self.color = color

      _Object.__init__(self, **kw)

   def plot(self):
      '''Plot the axis.  Usually called by the plot() function of the Plot
         class.'''
      ppgplot.pgsave()
      ppgplot.pgsch(self.fsize)
      _set_color(self.color)
      ppgplot.pgscf(self.font)
      ppgplot.pgsls(self.linestyle)
      ppgplot.pgslw(self.linewidth)
      ppgplot.pgaxis(self.x1, self.y1, self.x2, self.y2, self.options, 
                     self.v1, self.v2, self.step, self.nsub, self.dmajl,
                     self.dmajr, self.fmin, self.disp, self.orient)
      ppgplot.pgunsa()

class _Label(_Object):
   ''' A class used by the Plot class.  It defines a 
       label which will be drawn in the plot.'''
   def __init__(self, parent, x, y, string, angle=0, just=0, fsize=None, 
                font=None, color=None, width=None, vjust=0.0, bbox=None,
                reference='absolute', linewidth=None, **kw):
      '''The label initialization routine.  
      @param x:  the x world coordinate of the label
      @type x:   float
      @param y:  the y world coordinate of the label
      @type y:   float
      @param string:  the string to display
      @type string:  string
      @param angle:  The angle at which to draw the label (default: 0)
      @type angle:  float
      @param reference:  Are the coordinates 'absolute' world or 'relative' to the
                    viewport?  Default:  'absolute'
      @type reference:  string
      @param just:  The justification of the label (default: 0)
      @type just:  float
      @param fsize:  The character size to use (defaults to global size)
      @type fsize:  float
      @param font:  The font with which to draw the label 
                    (defaults to global font)
      @param color:  The color with which to draw the label 
                    (defaults to foreground)
      @param width:  The width (in world coordinates) of the label.  If the 
                     string is longer than this width, it is split into 
                     several lines.  Not available (yet) with nonzero angle 
                     (default None).  
      @type width:  float
      @param bbox:   Draw a bouding box around the label.  If a single
                     color string or integer, a box is drawn around the
                     label.  If a list of 2 color strings or integers, 
                     the first is used to draw the box and the second is
                     used to fill in the background (making it opaque).
                     (default None).
      @type bbox:  color or list of 2 colors
      @param vjust:  The vertical justification of the label, if it is split
                     over several lines.  If you imagine the bounding box to
                     have sides (dx and dy), then the anchor of the box is
                     lower left corner plus (dx*just, dy*vjust).  (default 0)
      @type vjust:  float
      '''
      self.parent = parent
      self.x = x
      self.y = y
      self.string = string
      self.just = just
      if fsize:  self.fsize = fsize
      if font:  self.font = font
      if linewidth:  self.linewidth = linewidth
      self.angle = angle
      if color:  self.color = color
      self.width = width
      self.vjust = vjust
      self.bbox = bbox
      if self.angle != 0:
         self.width = None
      self.reference = reference
      _Object.__init__(self, **kw)

   def plot(self):
      '''plots the label (usually called by the plot() fuction of its
          parent class.'''
      ppgplot.pgsave()
      ppgplot.pgsch(self.fsize)
      ppgplot.pgslw(self.linewidth)
      _set_color(self.color)
      ppgplot.pgscf(self.font)

      res = self._str_bounding_box()
      xwpc = (res[0][2] - res[0][0])/len(self.string.replace('\n',''))
      ywpc = res[1][1] - res[1][0]

      (xch,ych) = ppgplot.pgqcs(4)
      #  the padding around the legend text.
      self._ypad = 0.3*ych
      self._xpad = 0.3*xch


      if len(string.split(self.string,'\n')) > 1:
         # The string is already split up.  Use this as breaks and trust
         #  the user knows what she's doing.
         substr = string.split(self.string, '\n')
         substr.reverse()
      else:
         if self.width is not None:
            max_length = int(self.width/xwpc)
            if max_length < len(self.string):
               substr = tw.wrap(self.string,max_length)
               substr.reverse()
            else:
               substr = [self.string]
         else:
            substr = [self.string]

      # Here, figure out coordinates of the label.  If relative coordinates,
      #  make the transformation to absolute real.
      if self.reference == 'absolute':
         x0 = self.x
         y0 = self.y - self.vjust*len(substr)*ywpc
      else:
         x0 = self.parent.xmin + self.x*(self.parent.xmax - self.parent.xmin)
         y0 = self.parent.ymin + self.y*(self.parent.ymax - self.parent.ymin)
         y0 = y0 - self.vjust*len(substr)*ywpc

      i = 0
      if self.bbox is not None:
         if type(self.bbox) == types.ListType:
            if len(self.bbox) != 2:
               raise ValueError,"bbox must be a 2-tuple (foregroud,background)\
                     or a single color (foreground)"
            (fore,back) = self.bbox
         else:
            (fore,back) = (self.bbox, None)

         if self.width:
            w = self.width
         else:
            w = (res[0][2] - res[0][0])
         x1 = x0 + w
         y1 = y0 + len(substr)*ywpc
         ppgplot.pgsfs(1)

         if back is not None:
            _set_color(back)
            ppgplot.pgrect(x0-self.just*w-self._xpad,x1-self.just*w+self._xpad,
                        y0-self._ypad,y1)
         _set_color(fore)
         ppgplot.pgsfs(2)
         ppgplot.pgrect(x0-self.just*w-self._xpad,x1-self.just*w+self._xpad
               ,y0-self._ypad,y1)
      for str in substr:
         ppgplot.pgptxt(x0, y0+i*ywpc, self.angle, self.just, str)
         i = i + 1

      ppgplot.pgunsa()
      
   def _str_bounding_box(self):
      '''Returns the bounding box as two lists x and y coordinates for
      the single string (not broken up into several lines).'''
      ppgplot.pgsave()
      ppgplot.pgsch(self.fsize)
      ppgplot.pgscf(self.font)
      res = ppgplot.pgqtxt(self.x,self.y,self.angle,self.just,
            self.string.replace('\n',''))
      ppgplot.pgunsa()
      return(res)

class _Legend(_Object):
   '''A class used by the Plot class to make a legend.  We need a reference
   to the parent Plot class, so that we can access the data attributes.'''
   def __init__(self, parent, height=None, width=None, position='ur', dx=None,
         dy=None, bbox = None, fsize=None, font=None, color=None,
         linewidth=None, linelength=None, **kw):
      self.parent = parent
      self.height = height
      self.width = width
      self.position = position
      self.dx = dx
      self.dy = dy
      self.bbox = bbox
      if fsize:  self.fsize = fsize
      if font:  self.font = font
      if color:  self.color = color
      if linewidth:  self.linewidth = linewidth
      self.linelength = linelength
      _Object.__init__(self, **kw)

   def _pattern_length(self, linestyle, linewidth):
      '''Returns the length needed (in world coordinates) for a line of
      width 'linewidth' in order that the full pattern (dash, dash-dot, etc)
      is plotted.  If it doesn't matter (solid, dotted), returns largest.
      The factors below are for 3 dashes (style=2), dash-dot-dash-dot 
      (style=3) and dash-dot-dot-dot-dash (style=5).  These were determined
      empirically (didn't feel like wading through the fortran).'''
      if self.linelength is not None:
         # Override from user
         return(self.linelength)
      else:
         (xmin,xmax,ymin,ymax) = ppgplot.pgqwin()
         factors = {1:0.045, 2:0.045 , 3:0.0315, 4:0.045, 5:0.037}
         return(abs(xmax-xmin)*factors[linestyle]*Numeric.sqrt(linewidth))


   def _bounding_box(self, width, height):
      '''Returns (x1,x2,y1,y2), which is the bounding box of the legend,
      given the width and height of the text of the legend in world coords.'''
      (xmin,xmax,ymin,ymax) = ppgplot.pgqwin()
      (xch,ych) = ppgplot.pgqcs(4)
      #  the padding around the legend text.
      self._ypad = 0.3*ych
      self._xpad = 0.3*xch

      if not self.dx:  
         xborder = 0.02*(xmax - xmin)
      else: 
         xborder = self.dx
      if not self.dy:  
         yborder = 0.02*(ymax - ymin)
      else:
         yborder = self.dy

      # are we in the upper or lower?
      if self.position[0] == 'u':
         y0 = ymax - yborder - height - 2*self._ypad
         y1 = y0 + height + 2*self._ypad
      else:
         y0 = ymin + yborder
         y1 = y0 + height + 2*self._ypad

      if self.position[1] == 'r':
         x0 = xmax - xborder - 2*self._xpad - width
         x1 = x0 + width + 2*self._xpad
      else:
         x0 = xmin + xborder
         x1 = x0 + width + 2*self._xpad

      return(x0,x1,y0,y1)

   def _get_scale(self):
      xfactor = 0
      yfactor = 0
      #ppgplot.pgsave()
      ppgplot.pgscf(self.font)
      ppgplot.pgsch(self.fsize)

      # Character height with current font size
      (xch,ych) = ppgplot.pgqcs(4)
      # figure out initial dimensions of the legend
      height = 0.0
      width = 0.0
      for item in self._line_items:
         (xl,yl) = ppgplot.pglen(item.label,4)
         linelength = self._pattern_length(item.linestyle, item.linewidth)
         if xl + linelength > width:  width = xl + linelength
         height = height + ych
      for item in self._point_items:
         (xl,yl) = ppgplot.pglen(item.label,4)
         if xl + 3*xch > width:  width = xl + 3*xch
         height = height + ych
      if not self.width and not self.height:  return(1.0, width, height)

      # Now, figure out the font size to make the height/width fit, if needed.
      if self.height is not None:
         yfactor = self.height/height
         if not self.width:  xfactor = yfactor

      if self.width is not None:
         xfactor = self.width/width
         if not self.height:  yfactor = xfactor

      factor = min(xfactor,yfactor)
      height = height * factor
      width = width * factor
      #ppgplot.pgunsa()
      return(factor, width, height)

   def plot(self):
      if not self.fsize and not self.height:  
         fsize = self.parent.options['fsize']
      else:
         fsize = self.fsize
      # First, find out how many data sets are labeled
      self._line_items = []
      self._point_items = []
      for thing in self.parent.lines:
         if thing.label is not None:
            self._line_items.append(thing)
      for thing in self.parent.points:
         if thing.label is not None:
            self._point_items.append(thing)
      self._nitems = len(self._line_items + self._point_items)

      (scale, width, height) = self._get_scale()
      ppgplot.pgsch(self.fsize*scale)
      ppgplot.pgslw(self.linewidth)
      (xch, ych) = ppgplot.pgqcs(4)

      # Make a box around the legend.  Also, make it opaque background
      (x0,x1,y0,y1) = self._bounding_box(width,height)
      if self.bbox is not None:
         if type(self.bbox) == types.ListType or \
            type(self.bbox) == types.TupleType:
            if len(self.bbox) != 2:
               raise ValueError,"bbox must be a 2-tuple (foregroud,background)\
                     or a single color (foreground)"
            (fore,back) = self.bbox
         else:
            (fore,back) = (self.bbox, None)

         ppgplot.pgsave()
         if back is not None:
            ppgplot.pgsfs(1)
            _set_color(back)
            ppgplot.pgrect(x0,x1,y0,y1)

         _set_color(fore)
         ppgplot.pgsfs(2)
         ppgplot.pgrect(x0,x1,y0,y1)
         ppgplot.pgunsa()


      self._line_items.reverse()
      self._point_items.reverse()
      maxlinelength = 2*xch
      ll = xch
      for item in self._line_items:
         ll = self._pattern_length(item.linestyle, item.linewidth)
         if ll > maxlinelength:  maxlinelength = ll

      i = 0
      for item in self._line_items + self._point_items:
         ppgplot.pgptxt(x0 + self._xpad + ll + xch, y0 + self._ypad + i*ych,
               0.0, 0.0, item.label)
         i = i + 1

      i = 0
      for item in self._line_items:
         ppgplot.pgsave()
         ppgplot.pgslw(item.linewidth)
         ppgplot.pgsls(item.linestyle)
         _set_color(item.color)
         linelength = self._pattern_length(item.linestyle, item.linewidth)
         xxs = Numeric.array([x0 + self._xpad + (maxlinelength - linelength),
                x0 + self._xpad + maxlinelength])
         yys = Numeric.array([y0+self._ypad+i*ych,y0+self._ypad+i*ych])+0.3*ych
         ppgplot.pgline(xxs, yys)
         i = i + 1
         ppgplot.pgunsa()

      for item in self._point_items:
         ppgplot.pgsave()
         # First, fill in if needed:
         if item.fill:
            _set_color(item.fillcolor)
            ppgplot.pgslw(3)
            for size in Numeric.arange(self.fsize, 0.0, -0.05):
               ppgplot.pgsch(size)
               ppgplot.pgpt(Numeric.array([x0+self._xpad+xch]),
                     Numeric.array([y0+self._ypad+i*ych])+0.3*ych, item.symbol)
         ppgplot.pgsch(self.fsize)
         _set_color(item.color)
         ppgplot.pgpt(Numeric.array([x0+self._xpad+xch]), 
                      Numeric.array([y0+self._ypad+i*ych])+0.3*ych, item.symbol)
         ppgplot.pgunsa()
         i = i + 1
   
class _ErrorBar(_Object):
   '''A class used by the Plot class to store information about drawing
       errorbars'''
   def __init__(self, parent, x, y, dy1=None, dy2=None, dx1=None, dx2=None, length=1,
                linestyle=None, linewidth=None, color=None, **kw):
      '''Initialization function for ErrorBar.  If dy1 and dy2 are
         specified, you get uneven errorbars (unless dy1 = dy2).  If
         you only specify dy1, you get symmetric errorbars.  Same goes
         for dx1 and dx2
         @param x:  an array of x world coordinates for the errorbars
         @type x:   float array
         @param y:  an array of y world coordinates for the errorbars
         @type y:   float array
         @param dy1:  an array of upper limits for the y errorbars
         @type dy1:  float array
         @param dy2:  an array of lower limits for the y errorbars
         @type dy2:  float array
         @param dx1:  an array of upper limits for the x errorbars
         @type dx1:  float array
         @param dx2:  an array of lower limits for the x errorbars
         @type dx2:  float array
         @param length:  multiplier for default length of errorbar terminals
         @type length:  float
         @param linestyle:  Linestyle to use when drawing the errorbars 
                               (defaults to global value)
         @type linestyle:  integer
         @param linewidth:  Linewidth to use for drawing the errorbars
                               (defaults to global value)
         @type linewidth:  integer
         @param color:  The color use (defaults to global color)
         @type color:  integer or string
      '''
      self.parent = parent
      self.x = x
      self.y = y
      self.dy1 = dy1
      self.dy2 = dy2
      self.dx1 = dx1
      self.dx2 = dx2
      self.length = length
      if linestyle:  self.linestyle = linestyle
      if linewidth:  self.linewidth = linewidth
      if color:  self.color = color
      _Object.__init__(self, **kw)

   def range(self):
      '''Return the range needed to plot all the data.'''
      if self.dx1 is None:  
         dx1 = 0*self.x
         dx2 = 0*self.x
      else:
         dx1 = self.dx1
         if self.dx2 is None:  
            dx2 = dx1
         else:
            dx2 = self.dx2
      if self.dy1 is None:  
         dy1 = 0*self.y
         dy2 = 0*self.y
      else:
         dy1 = self.dy1
         if self.dy2 is None:  
            dy2 = dy1
         else:
            dy2 = self.dy2

      upperx = self.x + dx1
      lowerx = self.x - dx2
      uppery = self.y + dy1
      lowery = self.y - dy2
      return(([nummin(lowerx),nummax(upperx)],[nummin(lowery),nummax(uppery)]))

   def plot(self):
      '''Plot the errorbars.  Usually called from the parent class.'''
      ppgplot.pgsave()
      ppgplot.pgsls(self.linestyle)
      ppgplot.pgslw(self.linewidth)
      _set_color(self.color)
      if self.dx1 is not None and self.dx2 is not None:
         ppgplot.pgerrb(1, self.x, self.y, self.dx1, self.length)
         ppgplot.pgerrb(3, self.x, self.y, self.dx2, self.length)
      elif self.dx1 is not None:
         ppgplot.pgerrb(5, self.x, self.y, self.dx1, self.length)

      if self.dy1 is not None and self.dy2 is not None:
         ppgplot.pgerrb(2, self.x, self.y, self.dy1, self.length)
         ppgplot.pgerrb(4, self.x, self.y, self.dy2, self.length)
      elif self.dy1 is not None:
         ppgplot.pgerrb(6, self.x, self.y, self.dy1, self.length)
      ppgplot.pgunsa()
      

class _Point(_Object):
   '''Class used by Plot to store information about drawing points.
    '''
   def __init__(self, parent, x, y, symbol=None, size=None, color=None,
         label=None, linewidth=None, fill=0, fillcolor=None, **kw):
      '''Initialization for Point class.
      @param x:  an array of x world coordinates of the points
      @type x:   float array
      @param y:  an array of y world coordinates of the points
      @type y:   float array
      @param symbol:  Symbol to use when drawing the points
                         (default: 0)
      @type symbol:  integer
      @param size:  Size of the symbol.  (defaults to global value)
      @type size:  float
      @param fill:  Fill symbol with a color?  (default: 0)
      @type fill:  boolean
      @param color:  The color use (defaults to global color)
      @type color:  integer or string
      @param fillcolor:  The color with which to fill (defaults to global color)
      @type fillcolor:  integer or string
      @param label:  optional label for a lengend
      @type label:  string
      '''
      self.parent = parent
      self.x = x
      self.y = y
      if symbol:  self.symbol = symbol
      if size:  self.size = size
      if color:  self.color = color
      if linewidth:  self.linewidth = linewidth
      self.fill = fill
      if fillcolor is None:
         self.fillcolor = _default_color
      else:
         self.fillcolor = fillcolor
      self.label = label
      _Object.__init__(self, **kw)

   def plot(self):
      '''Plot the symbols.  Normall called by the parent class.'''
      ppgplot.pgsave()
      # Now, if we want to fill the symbol, do a bunch of smaller versions:
      if self.fill:
         _set_color(self.fillcolor)
         ppgplot.pgslw(3)
         for size in Numeric.arange(self.size, 0.0, -0.05):
            ppgplot.pgsch(size)
            ppgplot.pgpt(self.x, self.y, self.symbol)
      ppgplot.pgsch(self.size)
      ppgplot.pgslw(self.linewidth)
      _set_color(self.color)
      ppgplot.pgpt(self.x,self.y,self.symbol)
      
      ppgplot.pgunsa()

   def range(self):
      '''Get the plotting range such that all the symbols would be visible.'''
      return [_scalerange(self.x),_scalerange(self.y)]

class _Line(_Object):
   '''Class used by Plot and to store information about lines.'''
   def __init__(self, parent, x, y, linestyle=None, linewidth=None, color=None,
                label=None, **kw):
      '''Initialization function for Line.  
      @param x:  an array of x world coordinates for the line segments
      @type x:   float array
      @param y:  an array of y world coordinates for the line segments
      @type y:   float array
      @param linestyle:  Linestyle to use when drawing the line 
                         (defaults to global value)
      @type linestyle:  integer
      @param linewidth:  Linewidth to use for drawing the line
                         (defaults to global value)
      @type linewidth:  integer
      @param color:  The color use (defaults to global color)
      @type color:  integer or string
      @param label:  optional label for a legend
      @type label:  string.
      '''
      self.parent = parent
      self.x = x
      self.y = y
      self.parent = parent
      if linestyle:  self.linestyle = linestyle
      if linewidth:  self.linewidth = linewidth
      if color:  self.color = color
      self.label = label
      _Object.__init__(self, **kw)
   
   def plot(self):
      '''Plot the line.  Normall used by the parent class.'''
      ppgplot.pgsave()
      ppgplot.pgsls(self.linestyle)
      ppgplot.pgslw(self.linewidth)
      _set_color(self.color)
      ppgplot.pgline(self.x,self.y)
      ppgplot.pgunsa()
   def range(self):
      '''Get the plot range required to show all the line.'''
      return [_scalerange(self.x),_scalerange(self.y)]

class _Hist(_Object):
   '''Class used by Plot to make a Histogram.'''
   def __init__(self, parent, x, linestyle=None, linewidth=None, linecolor=1,
                fillcolor=None, fillstyle=1, low=None, high=None, nbin=10, 
                fill=1, outline=1, hatchangle=45, hatchsep=1.0, hatchphase=0.0,
                **kw):
      self.parent = parent
      self.x = Numeric.asarray(x)
      if linestyle is not None:  self.linestyle = linestyle
      if linewidth is not None:  self.linewidth = linewidth
      self.linecolor = linecolor
      self.fillstyle = fillstyle
      if fillcolor is not None:  
         self.fillcolor = fillcolor
      else:
         self.fillcolor = 1
      self.fill = fill
      self.outline = outline
      if low is not None:
         self.min = low
      else:
         self.min = nummin(self.x)
      if high is not None:
         self.max = high
      else:
         self.max = nummax(self.x)
      self.nbin = nbin
      self.style = fill
      self.hatchangle=hatchangle
      self.hatchsep=hatchsep
      self.hatchphase=hatchphase
      _Object.__init__(self, **kw)

   def plot(self):
      '''Plot the line.  Normall used by the parent class.'''
      self.bin()
      ppgplot.pgsave()
      if self.fill:
         _set_color(self.fillcolor)
         ppgplot.pgsfs(self.fillstyle)
         ppgplot.pgshs(self.hatchangle, self.hatchsep, self.hatchphase)
         ppgplot.pghist_s(self.x, self.nbin, 3, self.min, self.max)
         #ppgplot.pgbin(self.midpts, self.bins, 1)
      if self.outline:
         ppgplot.pgsls(self.linestyle)
         ppgplot.pgslw(self.linewidth)
         _set_color(self.linecolor)
         ppgplot.pgsfs(2)
         #ppgplot.pgbin(self.midpts, self.bins, 1)
         ppgplot.pghist_s(self.x, self.nbin, 3, self.min, self.max)
      ppgplot.pgunsa()
   def bin(self):
      # Bin the data
      self.bins = []
      delta = float(self.max - self.min)/self.nbin
      for i in range(self.nbin):
         self.bins.append(Numeric.sum(
            Numeric.greater_equal(self.x, self.min+i*delta)*\
            Numeric.less(self.x, self.min+(i+1)*delta)))
      self.bins = Numeric.array(self.bins)
      self.midpts = Numeric.arange(self.nbin)*delta + self.min + delta/2.0
   def range(self):
      '''Get the plot range required to show all the line.'''
      # pghist figures it out for itself, so we need to be sneaky and also
      # figure out the distribution.
      self.bin()
      return((self.min, self.max), (0, max(self.bins)))
         
class _Contour(_Object):
   '''Class used to plot a contour of data.'''
   def __init__(self, parent, data, contours=None, ncontours=10, cfunc=Linear,
                high=None, low=None, irange=None, fsize=None, font=None,
                jrange=None, xrange=None, yrange=None, color=None,
                linewidth=None, linestyle=None,
                clabc=_default_clab_color, clabi=_default_clab_int,
                clabm=_default_clab_min, tr = None, clab = None,
                fill=None, fillstyle=1, fillcolor=1, **kw):
      '''Initialization function for Contour.
      @param data: A 2-D array to contour
      @type data: 2D array
      @param contours:  '''
      self.parent = parent
      self.data = data
      if linewidth:  self.linewidth = linewidth
      if linestyle:  self.linestyle = linestyle
      if color:  self.color = color
      self.clab = clab
      self.clabc = clabc
      self.clabm = clabm
      self.clabi = clabi
      if fsize:  self.fsize = fsize
      if font:  self.font = font
      self.fill = fill
      self.fillstyle=fillstyle
      self.fillcolor=fillcolor
      # Make sure the data is a proper Numeric array.
      try:
         self.data = Numeric.asarray(self.data)
      except:
         raise TypeError, 'First argument must be a matrix'

      # Check to make sure it's a 2-d array:
      if len(self.data.shape) != 2:
         raise TypeError, 'First argument must be a 2-D matrix'

      # i is the number of columns and j the number of rows.  This is kinda
      #   backwards in Numeric.
      self.isize = self.data.shape[1]
      self.jsize = self.data.shape[0]

      if not irange:
         (self.imin, self.imax) = (0, self.isize - 1)
      else:
         (self.imin, self.imax) = irange

      if not jrange:
         (self.jmin, self.jmax) = (0, self.jsize - 1)
      else:
         (self.jmin, self.jmax) = jrange

      # Now figure out the contours.  Several ways to go:
      # 1) use the contours variable by default
      if contours is not None:
         self.contours = Numeric.asarray(contours)
      # 2) use the cfunc to generate the contours
      else:
         # First, figure out the max and min contour levels
         if high: 
            self.high = high 
         else: 
            self.high = nummax(Numeric.ravel(self.data))
         if low: 
            self.low = low
         else:
            self.low = nummin(Numeric.ravel(self.data))

         # Now that that's sorted out, we call the cfunc to generate the
         # contour levels:
         self.contours = Numeric.asarray(cfunc(self.low,self.high,ncontours))

      # Now figure out the transformation matrix
      # First, if xrange and yrange are given, assume a linear trans. with
      #  no rotation
      if xrange or yrange:
         if not xrange or not yrange:
            raise SyntaxError, "both xrange and yrange must be specified"
         self.tr = Numeric.asarray([xrange[0], 
                   float(xrange[1] - xrange[0])/self.isize,
                   0.0, yrange[0], 0.0, 
                   float(yrange[1] - yrange[0])/self.jsize])
      # Next, if tr is specified explicitly, use it:
      elif tr:
         self.tr = Numeric.asarray(tr)
      # Lastly, if the user gives nothing, assume the world coordinates are
      # just pixel coordinates
      else: 
         self.tr = Numeric.array([0.0, 1.0, 0.0, 0.0, 0.0, 1.0])

      _Object.__init__(self, **kw)

   def _pix2world(self, px, py):
      '''Retunr the world coordinates corresponding to pixel coordinates px
         and py.'''
      return (self.tr[0] + self.tr[1]*px + self.tr[2]*py,
              self.tr[3] + self.tr[4]*px + self.tr[5]*py)

   def range(self):
      '''Return the xrange and yrange that would allow all the data to be
         plotted.'''
      p1 = self._pix2world(0,0)
      p2 = self._pix2world(0,self.jsize)
      p3 = self._pix2world(self.isize, 0)
      p4 = self._pix2world(self.isize, self.jsize)
      return ([min(p1[0],p2[0],p3[0],p4[0]),max(p1[0],p2[0],p3[0],p4[0])],
              [min(p1[1],p2[1],p3[1],p4[1]),max(p1[1],p2[1],p3[1],p4[1])])

   def plot(self):
      '''Plot the contour.'''
      ppgplot.pgsave()
      # Choose the font size
      ppgplot.pgsch(self.fsize)
      # Choose the line style
      ppgplot.pgsls(self.linestyle)
      # Choose the line width
      ppgplot.pgslw(self.linewidth)
      # Choose the font
      ppgplot.pgscf(self.font)
      # Choose the line color for the contourlines

      # first do fill if requested:
      if self.fill is not None:
         if len(self.contours) < 1:
            raise AttributeError, "Error:  len(contours) must be > 1 to fill between them!"
         if type(self.fill) in [type([]), type(Numeric.array([])), type(())]:
            if len(self.fill) != len(self.contours) - 1:
               raise ValueError, "Error:  len(fill) must be one less than len(contours)"
            fills = self.fill
         else:
            # assume all contours are to be filled
            fills = [1]*(len(self.contours)-1)

         #check validity of fillcolor
         if type(self.fillcolor) in [type([]), type(Numeric.array([])), type(())]:
            if len(self.fillcolor) != len(fills):
               raise ValueError, "Error:  len(fillcolor) must be one less than len(contours)"
            colors = self.fillcolor
         elif type(self.fillcolor) in [type(""), type(1)]:
            colors = [self.fillcolor]*len(fills)
         else:
            raise TypeError, "Error:  fill color must be an int, string, or list"

         #check validity of fillstyle
         if type(self.fillstyle) in [type([]), type(Numeric.array([])), type(())]:
            if len(self.fillstyle) != len(fills):
               raise ValueError, "Error:  len(fillstyle) must be one less than len(contours)"
            styles = self.fillstyle
         elif type(self.fillstyle) is type(1):
            styles = [self.fillstyle]*len(fills)
         else:
            raise TypeError, "Error:  fill style must be an int, or list of ints"

         # All's good.  Make the fills
         for i in range(len(fills)):
            if not fills[i]: continue
            _set_color(colors[i])
            ppgplot.pgsfs(styles[i])
            ppgplot.pgconf(self.data, self.isize, self.jsize,
                  self.imin, self.imax, self.jmin,
                  self.jmax, self.contours[i], self.contours[i+1],
                  self.tr)

      # Now go ahead and do the contours
      _set_color(self.color)
      ppgplot.pgcont(self.data, 
                     self.isize, self.jsize, 
                     self.imin, self.imax, self.jmin, 
                     self.jmax, self.contours, len(self.contours),
                     self.tr)
      # Label the contours if requested
      if self.clab:
         # Choose the line color for the contourlines
         if type(self.clabc) == types.StringType:
            ppgplot.pgsci(_color_table[self.clabc])
         else:
            ppgplot.pgsci(self.clabc)
         for i in range(len(self.contours)):
            ppgplot.pgconl(self.data, self.isize, self.jsize, 
                    self.imin, self.imax, self.jmin, 
                    self.jmax, self.contours[i], self.tr, 
                    '%.2g' % self.contours[i], self.clabi, self.clabm)
      ppgplot.pgunsa()

class _Image(_Object):
   '''Class used to display an image representation of data.'''
   def __init__(self, parent, data, palette=_default_palette, 
                transfer=_default_transfer, irange=None, jrange=None,
                high=None, low=None, xrange=None, yrange=None, 
                tr = None, autocut=None, bright=0.5, contrast=1.0,
                **kw):
      self.parent = parent
      # Make sure the data is a proper Numeric array.
      try:
         self.data = Numeric.asarray(data)
      except:
         raise TypeError, 'First argument must be a matrix'

      # Check to make sure it's a 2-d array:
      if len(self.data.shape) != 2:
         raise TypeError, 'First argument must be a 2-D matrix'

      # i is the number of columns and j the number of rows.  This is kinda
      #   backwards in Numeric.
      self.isize = self.data.shape[1]
      self.jsize = self.data.shape[0]

      self.palette = palette
      self.transfer = transfer
      self.bright = bright
      self.contrast = contrast
      if high:
         self.high = high
      else:
         self.high = nummax(Numeric.ravel(data))

      if low:
         self.low = low
      else:
         self.low = nummin(Numeric.ravel(data))

      self.autocut_computed = 0
      if autocut:
         if autocut < 0 or autocut > 100:
            raise InputError, "Autocut must be between 0 and 100 (percent)"

      self.autocut = autocut

      # If the pixel ranges aren't specified, use the size of the array
      if not irange:
         (self.imin, self.imax) = (0, self.isize - 1)
      else:
         (self.imin, self.imax) = irange

      if not jrange:
         (self.jmin, self.jmax) = (0, self.jsize - 1)
      else:
         (self.jmin, self.jmax) = jrange

      # Now figure out the transformation matrix
      # First, if xrange and yrange are given, assume a linear trans. with
      #  no rotation
      if xrange or yrange:
         if not xrange or not yrange:
            raise SyntaxError, "both xrange and yrange must be specified"
         self.tr = Numeric.asarray([xrange[0], 
                   float(xrange[1] - xrange[0])/self.isize,
                   0.0, yrange[0], 0.0, 
                   float(yrange[1] - yrange[0])/self.jsize])
      # Next, if tr is specified explicitly, use it:
      elif tr:
         self.tr = Numeric.asarray(tr)
      # Lastly, if the user gives nothing, assume the world coordinates are
      # just pixel coordinates
      else: 
         self.tr = Numeric.array([0.0, 1.0, 0.0, 0.0, 0.0, 1.0])

      _Object.__init__(self, **kw)

   def __setattr__(self, param, value):
      if param == 'autocut':
         self.__dict__['autocut'] = value
         self.__dict__['autocut_computed'] = 0
      else:
         self.__dict__[param] = value

      
   def _pix2world(self, px, py):
      '''Retunr the world coordinates corresponding to pixel coordinates px
         and py.'''
      return (self.tr[0] + self.tr[1]*px + self.tr[2]*py,
              self.tr[3] + self.tr[4]*px + self.tr[5]*py)

   def range(self):
      '''Return the xrange and yrange that would allow all the data to be
         plotted.'''
      p1 = self._pix2world(0,0)
      p2 = self._pix2world(0,self.jsize)
      p3 = self._pix2world(self.isize, 0)
      p4 = self._pix2world(self.isize, self.jsize)
      return ([min(p1[0],p2[0],p3[0],p4[0]),max(p1[0],p2[0],p3[0],p4[0])],
              [min(p1[1],p2[1],p3[1],p4[1]),max(p1[1],p2[1],p3[1],p4[1])])

   def plot(self):
      '''Plot the image.'''
      # Set the color indices and the color table.  Note:  make sure that
      # we set a limit to the color index range to be 16 to max, otherwise
      # the pre-defined color indeces will be nuked.
      lo_col_ind, hi_col_ind = ppgplot.pgqcol()
      if hi_col_ind - lo_col_ind < 20:
         raise IOError, "This device has less than 20 colors, plotting an image won't look very nice... try another device"
      lo_col_ind = lo_col_ind + 16
      ppgplot.pgscir(lo_col_ind, hi_col_ind)
      _pgpalette.setpalette(self.palette,self.transfer)
      ppgplot.pgctab(_pgpalette.l,_pgpalette.r,_pgpalette.g,_pgpalette.b,
            len(_pgpalette.l), self.contrast, self.bright)
      if self.autocut:
         if not self.autocut_computed:
            (min,max) = ppgplot.autocuts(self.data, self.autocut)
            self.low = min
            self.high = max
            self.autocut_computed = 1
         else:
            (min,max) = (self.low, self.high)
      else:
         (min,max) = (self.low,self.high)
      tr = self.tr + Numeric.array([-self.tr[1]/2,0,0,-self.tr[5]/2,0,0])
      ppgplot.pgimag(self.data, 
                     self.isize, self.jsize,
                     self.imin, self.imax, self.jmin, self.jmax,
                     min,max, self.tr)  

   def change_rgb(self, palette=None, transfer=None, bright=0.5, contrast=1.0):
      '''Change the colormap settings.  No call to replot() is made, so if 
      nothing happens, it means the device doesn't support on-the-fly chaning of
      the color map.'''
      if palette is not None:  self.palette = palette
      if transfer is not None:  self.transfer = transfer
      self.bright = bright
      self.contrast = contrast
      _pgpalette.setpalette(self.palette, self.transfer)
      ppgplot.pgctab(_pgpalette.l, _pgpalette.r, _pgpalette.g, _pgpalette.b,
            len(_pgpalette.l), self.contrast, self.bright)

class _ColorPoint(_Object):
   '''Class used to display (x,y) points (like _Point), but each point is 
   colored differently, depending on the value of a third array (z).  The
   user specifies the color palette and transfer function (default linear).'''
   def __init__(self, parent, x, y, z, symbol=None, size=None, 
         palette=_default_palette, transfer=_default_transfer, linewidth=None,
         **kw):
      # Make sure the data is a proper Numeric array.
      self.parent = parent
      try:
         self.x = Numeric.asarray(x)
         self.y = Numeric.asarray(y)
         self.z = Numeric.asarray(z)
      except:
         raise TypeError, 'First 3 arguments must be arrays'

      if symbol:  self.symbol = symbol
      if size:  self.size = size
      if linewidth:  self.linewidth=linewidth
      self.palette = palette
      self.transfer = transfer

      _Object.__init__(self, **kw)

   def range(self):
      '''Return the xrange and yrange that would allow all the data to be
         plotted.'''
      return [_scalerange(self.x),_scalerange(self.y)]

   def plot(self):
      '''Plot the image.'''
      # Set the color indices and the color table.  Note:  make sure that
      # we set a limit to the color index range to be 16 to max, otherwise
      # the pre-defined color indeces will be nuked.
      lo_col_ind, hi_col_ind = ppgplot.pgqcol()
      lo_col_ind = lo_col_ind + 16
      ppgplot.pgscir(lo_col_ind, hi_col_ind)
      _pgpalette.setpalette(self.palette, self.transfer)
      ppgplot.pgctab(_pgpalette.l,_pgpalette.r,_pgpalette.g,_pgpalette.b)

      # Now, we go through and bin things up.  low is the lower end of the
      # bin, upper is the upper end of the bin
      ids = range(lo_col_ind, hi_col_ind)
      n = len(ids)

      #min = Numeric.minimum.reduce(self.z)
      #max = Numeric.maximum.reduce(self.z)
      min = nummin(Numeric.ravel(self.z))
      max = nummax(Numeric.ravel(self.z))
      dz = float(max - min)/n
      lower = min
      upper = min + dz
      ppgplot.pgsave()
      ppgplot.pgslw(self.linewidth)
      ppgplot.pgsch(self.size)
      for i in range(n):
         ppgplot.pgsci(ids[i])
         indexes = Numeric.less(self.z, upper) * Numeric.greater_equal( self.z ,lower)
         ppgplot.pgpt(Numeric.compress(indexes, self.x), 
                      Numeric.compress(indexes, self.y), self.symbol)
         lower = lower + dz
         upper = upper + dz
      
      ppgplot.pgunsa()

class Plot:
   '''Class to make a single 2-D plot.  Use this class to produce a single
      frame plot, with axes, lines, points, etc.  See constructor for
      options.

      Example:
      
      >>> plot = Pgplot.Plot(device='test.ps/PS', xrange=[0,1], yrange=[0,10])
      >>> plot.point(x,y)
      >>> plot.line(x,y)
      >>> plot.plot()
      >>> plot.close()
   '''

   def __init__(self, **kw):
      '''Initialization routine.  The Plot class has the following options
         which can be specified:
            - B{aspect}:  (I{float}) The aspect ratio (h/w).  Defaults to
              device's default
            - B{bottompad,toppad,rightpad,leftpad}: (I{float}) If you want
              extra padding on the top/bottom/right/left margins (maybe
              your labels are going off the viewport), specify these in units
              of the default character height.  Default: 0
            - B{color}:  (I{int/string}) Color with which objects are drawn
              (lines, symbols, etc) Defaults to foreground color.
            - B{device}:  (I{string}) Pgplot device to which the plot is sent.
              Default:  First available interactive device
            - B{flipxaxis}:  (I{bool}) Flip the xmax/xmin?.  Default:  0
            - B{flipyaxis}:  (I{bool}) Flip the ymax/ymin?.  Default:  0
            - B{font}:  (I{int}) Font with which to print labels.  Default:  1
            - B{fsize}:  (I{float}) Font size with which to print labels.
              Default:  1.0
            - B{id}:  (I{bool}) Print the user ID?  Default:  0
            - B{linestyle}: (I{int}) Line style with which to draw data lines.
              Default:  1
            - B{linewidth}: (I{int}) Line width with which to draw data lines.
              Default:  1
            - B{noleftlabel}: (I{bool}) Suppress first x-axis tick label?
              Default:  0
            - B{norightlabel}: (I{bool}) Suppress last x-axis tick label?
              Default:  0
            - B{notoplabel}: (I{bool}) Suppress last y-axis tick label?
              Default:  0
            - B{nobottomlabel}: (I{bool}) Suppress first y-axis tick label?
              Default:  0
            - B{size}: (I{float}) Default size of graph markers.
            - B{symbol}: (I{int}) Symbol with which to plot points.  Default:
              1
            - B{ticks}:  (I{string}) Draw ticks 'in', 'out', or 'both'?
              Default: 'in'
            - B{title}:  (I{string}) Title of the plot.  Default:  None.
            - B{width}:  (I{string}) Width with which to draw all other lines
              besides the data lines.  Default:  1
            - B{[x|y]axis}:  (I{bool}) Draw the [x|y] axis ([y|x]=0)?  Default:
              0
            - B{[x|y]format}: (I{string}) Format of the [x|y]-axis labels.  
              'hms' produces 'hh:mm:ss' format, 'dms' produces 'dd:mm:ss' and
              'dec' is the usual decimal format.  Default:  'dec'.
            - B{[x2|y2]format}: (I{string}) Format of the secondary 
              [x|y]-axis labels.  'hms' produces 'hh:mm:ss' format, 'dms' 
              produces 'dd:mm:ss' and 'dec' is the usual decimal format.  
              Default:  'dec'.
            - B{[x|y]grid}:  (I{int/string}) Whether to plot the grid lines 
              for the x or y axes and what color to use.  Set x/ygrid to
              the color (string or integer) you want plotted.  Default:
              no grid is drawn.
            - B{nl[xy]l} (I{boolean}) no leading [xy] label on the axis if
              using "hms" or "dms" format.  Default:  0
            - B{[x|y]log}:  (I{bool}) Use logarithmic [x|y] axis?  Default:  0
            - B{[x|y]2log}:  (I{bool}) Use logarithmic [x2|y2] axis? Default:  0
            - B{[x|y]range}: (I{2-tuple}) [x|y] range.  Default:  auto-range.
            - B{[x|y]2range}: (I{2-tuple}) [x|y] range of second axis.
              Default:  auto-range.
            - B{[x|y]label}: (I{string}) Label on [x|y] axis.  Default:  None
            - B{[x|y]2label}: (I{string}) Label on second [x|y] axis.  Default:
              None
            - B{[x|y]tick}: (I{float}) x- and y-tick intervals in world-
              coordinate units.  Default:0 (PGPLOT decides for you)
            - B{[x|y]2tick}: (I{float}) x- and y-tick intervals in world-
              coordinate units for second [x|y] axis.  Default:0 (PGPLOT 
              decides for you)
            - B{rotylab}: (I{boolena}) Rotate y-axis numeric labels.  Default: 0
      ''' 
      # this is a dictionary of options that can be set to affect plot 
      #  attributes
      self.__dict__['options'] = {
         'aspect':None, 'color':_default_color, 'device':_default_device,
         'font':_default_font, 'fsize':_default_fsize, 'id':0,
         'size':None, 'linestyle':_default_linestyle, 'linewidth':_default_linewidth,
         'symbol':_default_symbol, 'ticks':'in', 'title':None,
         'width':_default_lw, 'xaxis':None, 'xlog':0, 'x2log':0, 'xrange':None,
         'x2range':None, 'yaxis':None, 'yrange':None, 'y2range':None,
         'xlabel':None, 'ylabel':None, 'x2label':None, 'y2label':None,
         'labelcolor':_default_color, 
         'ylog':0, 'y2log':0, 'xformat':'auto', 'yformat':'auto', 
         'x2format':'dec', 'y2format':'dec',
         'xgrid':None, 'ygrid':None, 
         'noleftlabel':0, 'norightlabel':0,'notoplabel':0, 'nobottomlabel':0,
         'flipxaxis':0, 'flipyaxis':0, 'toppad':None, 'bottompad':None,
         'leftpad':None, 'rightpad':None, 'rotylab':0, 'xtick':0.0, 'ytick':0.0,
         'x2tick':0.0, 'y2tick':0.0
      }
      self._id = None       # Pgplot device ID
      self._own_dev = 0     # are we responsible for closing the device?
      #self._panel_num_i = None # If we are in a Panel, which panel are we?
      #self._panel_num_j = None # If we are in a Panel, which panel are we?
      self._panel_num = None   # If we are in a Panel, which panel are we?
      self.parent = None       # If we are in a panel, this is the instance of the panel
      
      self.lines = []
      self.points = []
      self.errors = []
      self.labels = []
      self.axes = []
      self.contours = []
      self.images = []
      self.hists = []
      self.colorpoints = []
      self.the_legend = None    # We'd never need more than one, right?
 
      for key in kw.keys():
         if key in self.options.keys():
            self.options[key] = kw[key]
         else:
            raise AttributeError, "Plot does not have attribute %s" % key

   def __del__(self):
      # ensure the device is closed
      self.close()


   def __str__(self):
      return "Instace of the pygplot Plot class"

   def __setattr__(self, name, value):
      if name not in self.options.keys():
         self.__dict__[name] = value
      else:
         self.options[name] = value

   def __getattr__(self, name):
      if 'options' in self.__dict__:
         if name in self.__dict__['options']:
            return(self.__dict__['options'][name])
         else:
            raise AttributeError, "Error:  attribute %s not defined" % name

   def point(self, x, y, symbol=None, size=None, color=None, label=None,
         linewidth=None, fill=0, fillcolor=None, **kw):
      '''Add a data set that will be drawn with symbols.
         @param x: array of x world coordinates
         @type x:  numarray
         @param y: array of y world coordinates
         @type y:  numarray
         @param symbol:  symbol to use (defaults to parent class default)
         @type symbol: int
         @param size:  size of symbols
         @type size:  float
         @param color:  color of the symbols
         @type color:  int/string
         @param linewidth:  width with which to draw the symbol
         @type linewidth:  int
         @param fill:  fill the symbol with solid color?  (default:  0)
         @type fill:  boolean
         @param fillcolor:  color with which to fill the symbol
         @type fillcolor:  int/string
         @param label:  optional label to be used in a legend
         @type label:  string
      '''
      x = Numeric.asarray(x)
      y = Numeric.asarray(y)
      new_point = _Point(self, x,y,symbol,size,color,label,linewidth,fill,
            fillcolor, **kw)
      self.points.append(new_point)
      return new_point

   def color_point(self, x, y, z, symbol=None, size=None, 
                   palette=_default_palette, transfer=_default_transfer,
                   linewidth=None, **kw):
      '''Add a data set that will be drawn with symbols.  Unlike the usual
         point, here the value of z is used to determine an intensity of
         each point and, based on the chosen palette, this is drawn with
         differen colors.
         @param x: array of x world coordinates
         @type x:  numarray
         @param y: array of y world coordinates
         @type y:  numarray
         @param z: array of intensities for the x,y points
         @type z:  numarray
         @param symbol:  symbol to use (defaults to parent class default)
         @type symbol: int
         @param size:  size of symbols
         @type size:  float
         @param palette:  color palette to use.  See list_palettes().
         @type palette:  string
         @param transfer:  transfer function (intensity) to use.  See 
                          list_palettes().
         @type transfer:  string
      '''
      x = Numeric.asarray(x)
      y = Numeric.asarray(y)
      z = Numeric.asarray(z)
      new_colorpoint = _ColorPoint(self, x,y,z,symbol,size,palette,transfer,
            linewidth, **kw)
      self.colorpoints.append(new_colorpoint)
      return new_colorpoint
      
   def axis (self, x1, y1, x2, y2, v1, v2, log=0, label=0, dec=0, exp=0,
                step=0, nsub=0, dmajl=0.5, dmajr=0.5, fmin=0.5, disp=0.3,
                orient=0.0, fsize=None, font=None, linestyle=None, 
                linewidth=None, color=None, **kw):
      '''Add an alternate axis to the plot.  This does not alter the viewport
         nor change the way points are plotted;  it is merely a convenient way
         to add another axis.
         @param x1: x world coordinate of the beginning of the axis
         @type x1:  float
         @param y1: y world coordinate of the beginning of the axis
         @type y1:  float
         @param x2: x world coordinate of the end of the axis
         @type x2:  float
         @param y2: y world coordinate of the end of the axis
         @type y2:  float
         @param v1: value of the axis at (x1,y1)
         @type v1:  float
         @param v2: x value of the axis at (x2,y2)
         @type v2:  float
         @param step:  major tick spacing along the axis.
         @type step: float
         @param nsub:  number of minor ticks per major tick interval
         @type nsub:  int
         @param dmajl:  length of the major tick to the left of the axis in
                        units of character height
         @type dmajl:  float
         @param dmajr:  length of the major tick to the right of the axis in
                        units of character height
         @type dmajr:  float
         @param fmin:  length of the minor ticks as fraction of the major tick.
         @type fmin:  float
         @param disp:  displacement of numeric labels away from the axis.  In
                       units of character height.
         @type disp:  float
         @param orient:  angle at which to print labels, relative to the axis.
                         (default 0.0)
         @type orient:  float
         @param linestyle:  line style to use (defaults to the Plot's global 
                            value)
         @type linestyle: int
         @param linewidth:  line width to use (defaults to the Plot's global 
                            value)
         @type linewidth:  int
         @param fsize:  font size to use (defaults to the Plot's global value)
         @type fsize: float
         @param font: font with which to print the label (defaults to the 
                     Plot's global value)
         @type font: int
         @param color:  Color with which to print the label (defaults to the
                     Plot's global value)
         @type color: string/int'''
      options = ''
      if log != 0:  options = options + "L"
      if label != 0:  options = options + "N"
      if dec != 0: options = options + "1"
      if exp != 0: options = options + "2"
      new_axis = _Axis(self, x1,y1,x2,y2,v1,v2,options,step,nsub,dmajl,dmajr,
                         fmin,disp,orient,fsize,font,linestyle,linewidth, color,
                         **kw)
      self.axes.append(new_axis)
      return new_axis

   def circle(self, x, y, rad, linestyle=None, color=None, **kw):
      '''Draw a circle centered on (x,y) with radius r, all in world 
      coordinates.'''
      angles = Numeric.arange(0, 2*Numeric.pi, 0.02)
      xs = rad*Numeric.cos(angles) + x
      ys = rad*Numeric.sin(angles) + y
      return (self.line(xs,ys,linestyle=linestyle, color=color, **kw))

   def ellipse(self, x, y, major, minor, pa, linestyle=None, color=None,n=100,
         **kw):
      '''Draw an ellipse centered on (x,y) with major axis and minor axis,
      rotated by pa (degrees).'''
      angles = Numeric.arange(n+1)/float(n)*2*Numeric.pi
      #angles = Numeric.arange(0, 2*Numeric.pi+0.00001, 0.02)
      cosP = Numeric.cos(pa*Numeric.pi/180.0)
      sinP = Numeric.sin(pa*Numeric.pi/180.0)
      cosA = Numeric.cos(angles)
      sinA = Numeric.sin(angles)
      denom = Numeric.power(minor*cosA,2) + Numeric.power(major*sinA, 2)
      R = Numeric.sqrt(minor*major*minor*major/denom)
      Xell = R*cosA
      Yell = R*sinA
      xs = Xell*cosP-Yell*sinP + x
      ys = Xell*sinP+Yell*cosP + y
      return(self.line(xs, ys, linestyle=linestyle, color=color, **kw))

   def box(self, xll, yll, xur, yur, linestyle=None, color=None, **kw):
      '''Draw a box with lower left corner (xll,yll) and upper-right corder
      (xur,yur) in world coordinates.'''
      return(self.line([xll,xll,xur,xu,xll],[yll,yur,yur,yll,yll],
         linestyle=linestyle, color=color, **kw))

   def label(self, x, y, string, angle=0, just=0, fsize=None, font=None, 
             color=None, width=None, vjust=0, bbox=None, reference='absolute',
             linewidth=None, **kw):
      '''Add a label to the plot.
         @param x:  the x world coordinate of the label
         @type x:   float
         @param y:  the y world coordinate of the label
         @type y:   float
         @param string:  the string to display
         @type string:  string
         @param angle:  The angle at which to draw the label (default: 0)
         @type angle:  float
         @param reference:  Are the coordinates 'absolute' world or 'relative' to the
                       viewport?  Default:  'absolute'
         @type reference:  string
         @param just:  The justification of the label (default: 0)
         @type just:  float
         @param fsize:  The character size to use (defaults to global size)
         @type fsize:  float
         @param font:  The font with which to draw the label 
                       (defaults to global font)
         @param color:  The color with which to draw the label 
                       (defaults to foreground)
         @param width:  The width (in world coordinates) of the label.  If the 
                        string is longer than this width, it is split into 
                        several lines.  Not available (yet) with nonzero angle 
                        (default None).  
         @type width:  float
         @param bbox:   Draw a bouding box around the label.  If a single
                        color string or integer, a box is drawn around the
                        label.  If a list of 2 color strings or integers, 
                        the first is used to draw the box and the second is
                        used to fill in the background (making it opaque).
                        (default None).
         @type bbox:  color or list of 2 colors
         @param vjust:  The vertical justification of the label, if it is split
                        over several lines.  If you imagine the bounding box to
                        have sides (dx and dy), then the anchor of the box is
                        lower left corner plus (dx*just, dy*vjust).  (default 0)
         @type vjust:  float
      '''   
      new_label = _Label(self, x,y,string,angle,just,fsize,font,color,width,vjust,
            bbox,reference, linewidth, **kw)
      self.labels.append(new_label)
      return new_label
      
   def line(self, x, y, linestyle=None, linewidth=None, color=None, label=None,
         **kw):
      '''Add a data set that are connected with lines
         @param x: array of x world coordinates
         @type x:  numarray
         @param y: array of y world coordinates
         @type y:  numarray
         @param linestyle:  Linestyle to use when drawing the line 
                            (defaults to parent class default)
         @type linestyle:  int
         @param linewidth:  Linewidth to use for drawing the line
                            (defaults to parent class default) 
         @type linewidth:  int
         @param color:  The color use (defaults to parent class global)
         @type color:  int/string
         @param label:  optional label to be used in a legend
         @type label:  string
      '''
      x = Numeric.asarray(x)
      y = Numeric.asarray(y)
      new_line = _Line(self, x,y,linestyle, linewidth, color, label, **kw)
      self.lines.append(new_line)
      return new_line

   def error(self, x, y, dx1=None, dy1=None, dx2=None, dy2=None, linestyle=None,
             length=1, linewidth=None, color=None, **kw):
      '''Add error bars to the data points.  By specifying one or two
      error limits, you get symmetric or non-symmetric error bars,
      respectively.

      @param x:  Array of x-values (midpoints)
      @type x:   float array
      @param y:  Array of y-values (midpoints)
      @type y:   float array
      @param dx1:  Array of x error lengths.  If dx2 is not specified,
                   errorbars are drawn at +/- dx1, otherwise it is an
                   upper error.
      @type dx1: float array
      @param dy1:  Same as dx1 but an upper limit in the y-direction
      @type dy1: float array
      @param dx2: Array of lower error limits in the x-direction
      @type dx2: float array
      @param dy2: Array of lower error limits in the y-direction
      @type dy2: float array
      @param length:  length of the end terminals as fraction of default
      @type length:  float'''
      x = Numeric.asarray(x)
      y = Numeric.asarray(y)
      if dx1 is None and dy1 is None:
         raise AttributeError, 'must have at least dx1 or dy1 specified'
      if dx1 is not None: dx1 = Numeric.asarray(dx1)
      if dx2 is not None: dx2 = Numeric.asarray(dx2)
      if dy1 is not None: dy1 = Numeric.asarray(dy1)
      if dy2 is not None: dy2 = Numeric.asarray(dy2)
      new_error = _ErrorBar(self,x,y,dy1,dy2,dx1,dx2,length,linestyle, linewidth, 
            color, **kw)
      self.errors.append(new_error)
      return new_error

   def contour(self, data, **kw):
      '''Contour plot data.  The only required argument is a 2 dimensional
         array.  The rest of the arguments are as follows:
            - B{clab}: (I{bool}):  Draw contour labels?  Default: 0
            - B{clabc}: (I{int/string}) Color for the contour labels.  Defaults
                        to foreground color.
            - B{clabi}: (I{int}) Spacing along the contour between labels
                        in grid cells.
            - B{clabm}: (I{float}) Contours that cross less than clabm grid
                        cells will not be labeled.
            - B{color}:  (I{int/string}) Color of the contours.
            - B{contours}: (I{numarray}) The contour levels to draw.  If not
              specified, then they will be generated automatically based on
              'low', 'high', 'ncontours', and cfunc (see below)
            - B{cfunc}: (I{function}) A function that returns contour levels.
              The function will be run by plot command automatically if
              the contours aren't specified explicitly.  The function takes
              3 arguments:  the value of the low contour, the value of the
              high contour and the number of contours requested.  The default
              is Linear(), a function that gives ncontour contours equally
              spaced between low and high.
            - B{font}:  (I{int}) Font with which to print labels.  Default: 
              inherited from the Plot class.
            - B{fsize}:  (I{float}) Font size with which to print labels.
              Default:  inherited from the Plot class.
            - B{fill}:  (I{int}/I{array})fill the contours?  If not None, fill 
                       between contours.  If a list, then if list[i] is true, 
                       fill between contours i and i+1.
            - B{fillcolor}: (I{color}/I{list}) color(s) to use for ith fill.
            - B{fillstyle}: (I{int}/I{list}) style(s) to use for ith fill.
            - B{high}: (I{float}) The largest pixel value to contour.
            - B{linestyle}: (I{int}) Line style with which to draw contours.
              Default:  inherited from Plot class.
            - B{linewidth}: (I{int}) Line width with which to draw contours.
              Default:  inherited from Plot class.
            - B{low}: (I{float}) Lowest pixel value to contour.
            - B{ncontours}: (I{int}) Number of contours to plot (default 10).
            - B{tr}: (I{6-tuple}) The transformation matrix from pixel
              coordinates to world coordinates.  If px and py are the pixel
              coordinates, then the world coordinates x and y are given by:
              x = tr[0] + tr[1]*px + tr[2]*py
              y = tr[3] + tr[4]*px + tr[5]*py
              This option overrides [x|y]range.
            - B{[x|y]range}: (I{2-tuple}) [x|y] range.  The range of the pixel
              data.  This is used to figure out the pixel-to-world coordinate
              transformaton (see tr above).  Default:  auto-range.'''

      # data = Numeric.asarray(data)
      new_contour = apply(_Contour,[self, data],kw)
      self.contours.append(new_contour)
      return new_contour

   def image(self, data, **kw):
      '''Plot a grey-scale or color image representation of data on the
         background of the plot.  The only required argument is data, a
         2 dimensional array.  The other options are as follows:
            - B{autocut}: (I{float})  If specified, the B{low} and B{high}
              will be chosen automatically so that B{autocut} percent of the
              data is included.
            - B{high}: (I{float}) Pixels with values higher than 'high' are
              drawn with the maxinum value in the color table.  Default value
              is the highest value in the data.
            - B{low}: (I{float}) Pixel values lower than 'low' are drawn with
              the lowest color in the color table.  Default value is the lowest
              pixel value in the data.
            - B{palette}: (I{string}) The color palette to use.  Use the
              function list_palettes() to see a list of available palettes.
              If palette is not specified, the default grey-scale will be used.
            - B{transfer}: (I{string}) The transfer function (intensity) to
              use for the color palette (default is "real" i.e., linear).
            - B{tr}: (I{6-tuple}) The transformation matrix from pixel
              coordinates to world coordinates.  If px and py are the pixel
              coordinates, then the world coordinates x and y are given by:
              x = tr[0] + tr[1]*px + tr[2]*py
              y = tr[3] + tr[4]*px + tr[5]*py
              This option overrides [x|y]range.
            - B{[x|y]range}: (I{2-tuple}) [x|y] range.  The range of the pixel
              data.  This is used to figure out the pixel-to-world coordinate
              transformaton (see tr above).  Default:  auto-range.
      ''' 
      #data = Numeric.asarray(data)
      new_image = apply(_Image,[self,data],kw)
      self.images.append(new_image)
      return new_image

   def histogram(self, data, **kw):
      '''Plot a histogram of unbinned data.  The bins are computed automatically
         by the routine from the input data.  The other optional arguments are
            - B{data}:  (I{float}) Data to bin.
            - B{low}:  (I{float}) Minimum bin value.  Data < bin is not 
                        counted.  Default:  min of data
            - B{high}:  (I{float}) Maximum bin value.  Data > bin is not 
                        counted.  Default:  max of data
            - B{nbin}:  (I{int}) Number of bins.  Default: 10
            - B{outline}: (I{bool}) Outline the histogram boxes?  Default: 1
            - B{fill}: (I{bool}) Fill the histogram boxes?  Default: 1
            - B{linestyle}:  (I{int}) Linestyle to use when drawing the boxes
                            (defaults to global value)
            - B{linewidth}:  (I{int}) Linewidth to use for drawing the boxes
                               (defaults to global value)
            - B{linecolor}:  (I{string,int})  The color use to outline the
                           histogram boxes (defaults to global color)
            - B{fillstyle}:  (I{int}) Which fill style to use:  1- solid 2- 
                         outline 3- hatch 4- cross-hatch.  Default:  1
            - B{fillcolor}:  (I{string,int})  The color with which to fill
                         the histogram box (defaults to global color)
            - B{hatchangle}:  (I{float}) The angle at which to draw the hatches
                         (default 45 deg.)
            - B{hatchsep}:  (I{float}) The separation between hatches in units of
                         1% the width or hight of plot (whichever is biggest).
                         default:  1.0
            - B{hatchphase}:  (I{float}) The phase of the hatchmarks.  Same phase
                         means adjacent bars have contiguous lines.
      '''
      data = Numeric.asarray(data)
      new_hist = apply(_Hist, [self, data], kw)
      self.hists.append(new_hist)
      return new_hist

   def legend(self, height=None, width=None, position='ur', dx=None, dy=None,
         bbox = None, fsize=None, font=None, color=None, linewidth=None,
         linelength=None, **kw):
      '''Draw a legend for the data.  You can control the height of the
         bounding box in world coordinates if you want to, but that will
         over-ride the font size (fsize has no effect).  The position
         can be ur (upper right), lr (lower right), ul (upper left), or
         ll (lower left).  dx and dy allow you to offset the legend from
         the axis (positive number brings it inside the graph).  If bbox
         is not None, a bounding box will be drawn.
           - B{height}:  max height of the legend in world coordinates (default:
                         as high as needed given the fsize)
           - B{widht}:   max width of the lengend in world coordinates (default:
                         as wide as needed given fsize
           - B{position}: vertical and horizontal alignment (ul, ur, ll, lr)
           - B{dx,dy}:   horizontal and vertical offsets from the axis
           - B{bbox}:    boolean:  draw a bounding box around the legend and
                         over-write what's underneath?
           - B{fsize}:   font size.  This over-rides height and width
           - B{font}:    font to use
           - B{linewidth}:  width for text objects
           - B{linelength}:  length of the lines in the legend (world coords) 
           - B{color}:   color for the text'''
      new_legend = _Legend(self, height, width, position, dx, dy, bbox,fsize,
            font, color, linewidth, linelength, **kw)
      self.the_legend = new_legend
      return new_legend

   def _set_viewport(self, just, compute_only=0):
      '''This sets the viewport according to the justification parameter
      (just).  It makes sure that if we are using Panels, the viewports
      of each sub-panel are snug.  We're also going to make sure there
      is space for titles (if any), x2labels (if any) and secondary
      ranges (if any)'''
      (vx1,vx2,vy1,vy2) = (0,1,0,1)  ;#  The whole surface
      if isinstance(self.parent, Panel) and not compute_only:
         self.parent.layout.pgsvp(vx1,vx2,vy1,vy2)
         return


      # Get an idea of how much space is needed for margins
      ppgplot.pgsave()
      ppgplot.pgscf(self.options['font'])
      ppgplot.pgsch(self.options['fsize'])
      (xch,ych) = ppgplot.pgqcs(0)
      ppgplot.pgunsa()
      # take off space for axis labels and a bit for the other margins:
      vy1 = vy1 + 2*ych
      vx1 = vx1 + 2*ych
      vy2 = vy2 - ych
      vx2 = vx2 - ych

      #  Lop off bits of the viewport to make space for each thing we need
      if self.options['title']:  vy2 = vy2 - 2*ych
      if self.options['xlabel']:  vy1 = vy1 + 2*ych
      if self.options['ylabel']:  vx1 = vx1 + 2*ych
      if self.options['x2range']:  vy2 = vy2 - 2*ych
      if self.options['y2range']:  vx2 = vx2 - 2*ych
      if self.options['x2label']:  vy2 = vy2 - 2*ych
      if self.options['y2label']:  vx2 = vx2 - 2*ych

      # Finally, if the user asks for more space, honour it.
      if self.options['leftpad']:  vx1 = vx1 + self.options['leftpad']*ych
      if self.options['rightpad']:  vx2 = vx2 - self.options['rightpad']*ych
      if self.options['toppad']:  vy2 = vy2 - self.options['toppad']*ych
      if self.options['bottompad']:  vy1 = vy1 + self.options['bottompad']*ych

      # Now, if there is a justification (for the Panel multiplots), then we
      #  need to reset some of the margins to zero.
      if just in [1,2,4,5,7,8,10,15]: vx2 = 1.0
      if just in [2,3,5,6,8,9,11,15]: vx1 = 0.0
      if just in [1,2,3,4,5,6,12,14]: vy1 = 0.0
      if just in [4,5,6,7,8,9,13,14]: vy2 = 1.0

      if compute_only:  return(vx1,vx2,vy1,vy2)
      if self.parent is not None:
         self.parent.layout.pgsvp(vx1,vx2,vy1,vy2)
      else:
         ppgplot.pgsvp(vx1,vx2,vy1,vy2)

   def _set_window(self, rangex, rangey, just, other=0):
      '''This internal routine sets up the viewport and window for the
         plot or sub-panel.  It takes care of figuring out which range
         to use and whether or not we need to elimitate the first labels
         on the axes (if using Panel, or if the user requests it).  The
         (other) parameter tells us if we are dealing with the other
         axes (rhs y axis or top x-axis)'''

      deltax = (rangex[1]-rangex[0])*0.0001
      deltay = (rangey[1]-rangey[0])*0.0001
      # make the ranges "not nice" numbers to stop first labels from
      #  being printed, if needed or requested
      if other:
         if (self.options['norightlabel'] or just in [1,2,10,15]):  
            rangex[1] = rangex[1] - deltax
         if (self.options['noleftlabel'] or just in [2,3,15,11]):
            rangex[0] = rangex[0] + deltax
         if (self.options['nobottomlabel'] or just in [3,6,12,14]):
            rangey[0] = rangey[0] + deltay
         if (self.options['notoplabel'] or just in [6,9,13,14]):
            rangey[1] = rangey[1] - deltay
      else:
         if (self.options['norightlabel'] or just in [7,8,10,15]):  
            rangex[1] = rangex[1] - deltax
         if (self.options['noleftlabel'] or just in [8,9,15,11]):
            rangex[0] = rangex[0] + deltax
         if (self.options['nobottomlabel'] or just in [1,4,12,14]):
            rangey[0] = rangey[0] + deltay
         if (self.options['notoplabel'] or just in [4,7,13,14]):
            rangey[1] = rangey[1] - deltay
      ppgplot.pgswin(rangex[0], rangex[1], rangey[0], rangey[1])

   def get_range(self):
      '''Get the current window range of the graph.'''
      return(ppgplot.pgqwin())
      
   def _draw_box(self, rangex, rangey, just):
      '''Simply draws the exterior box, axes, and ticks.'''

      ppgplot.pgsave()
      ppgplot.pgsch(self.options['fsize'])
      ppgplot.pgscf(self.options['font'])
      ppgplot.pgsls(self.options['linestyle'])
      ppgplot.pgslw(self.options['width'])
      _set_color(self.options['color'])

      # Decide if secondary (top and right) edges of boxes have different
      # ranges.
      if self.options['x2range']:
         otherxaxis=1
         rangex2 = self.options['x2range']
      else:
         rangex2=rangex
         otherxaxis=0
      if self.options['y2range']:
         otheryaxis=1
         rangey2=self.options['y2range']
      else:
         rangey2=rangey
         otheryaxis=0

      if self.options['ticks']=='in': 
         tick_env = ""
      elif self.options['ticks']=='both': 
         tick_env = "P"
      else:
         tick_env = "I"

      # First do the top and right (secondary) axes.  If they have a different
      #  scale, we need to first set the window
      if otherxaxis or otheryaxis:
         self._set_window(rangex2, rangey2, just, 1)
      else:
         self._set_window(rangex, rangey, just,0)

      # top axis
      xenv = "CST" + tick_env
      if self.options['x2log']: xenv = xenv + 'L'
      if just in [0,1,2,3,10,11,12,15]:  
         if self.options['x2range']: xenv = xenv + "M"
         if self.options['noleftlabel']: xenv = xenv + 'F'
         if self.options['x2format'] == 'hms':
            xenv = xenv + 'ZYXHO'
         elif self.options['x2format'] == 'dms':
            xenv = xenv + 'ZYDO'
      ppgplot.pgtbox(xenv, self.options['x2tick'], 0, "", 0.0, 0)

      #right axis
      yenv = "CST" + tick_env
      if self.options['y2log']: yenv = yenv + 'L'
      if self.options['rotylab']: yenv = yenv + 'V'
      if just in [0,3,6,9,11,12,13,14]:  
         if self.options['y2range']: yenv = yenv + "M"
         if self.options['nobottomlabel']: yenv = yenv + 'F'
         if self.options['y2format'] == 'hms':
            yenv = yenv + 'ZYXHO'
         elif self.options['y2format'] == 'dms':
            yenv = yenv + 'ZYDO'
      ppgplot.pgtbox("", 0.0, 0, yenv, self.options['y2tick'], 0)

      # Now that we've done the top and right axes, set the window to the
      #  righ ranges, if needed
      if otherxaxis or otheryaxis:
         self._set_window(rangex, rangey, just, 0)

      #bottom axis
      xenv = "BST" + tick_env
      if self.options['xlog']:  xenv = xenv + "L"
      if just in [0,7,8,9,10,11,13,15]: 
         xenv = xenv + "N"
         if self.options['xformat'] == 'hms': 
            xenv = xenv + "ZYXH0"
         elif self.options['xformat'] == 'dms':
            xenv = xenv + "ZYD0"
         elif self.options['xformat'] == 'dec':
            xenv = xenv + "1"
         elif self.options['xformat'] == 'exp':
            xenv = xenv + "2"
      ppgplot.pgtbox(xenv, self.options['xtick'], 0, "", 0.0, 0)

      # left axis
      yenv = "BST" + tick_env
      if self.options['ylog']: yenv = yenv + "L"
      if self.options['rotylab']: yenv = yenv + 'V'
      if just in [0,1,4,7,10,12,13,14]: 
         yenv = yenv + "N"
         if self.options['yformat'] == 'hms': 
            yenv = yenv + "ZYXH0"
         elif self.options['yformat'] == 'dms':
            yenv = yenv + "ZYD0"
         elif self.options['yformat'] == 'dec':
            yenv = yenv + "1"
         elif self.options['yformat'] == 'exp':
            yenv = yenv + "2"
      ppgplot.pgtbox("", 0.0, 0, yenv, self.options['ytick'], 0)

      if self.options['xgrid'] is not None:  
         _set_color(self.options['xgrid']); ppgplot.pgtbox('G',0,0,'',0,0)
      if self.options['ygrid'] is not None:  
         _set_color(self.options['ygrid']); ppgplot.pgtbox('',0,0,'G',0,0)
      if self.options['xaxis'] is not None:  
         _set_color(self.options['xaxis']); ppgplot.pgtbox('ATS',0,0,'',0,0)
      if self.options['yaxis'] is not None:  
         _set_color(self.options['yaxis']); ppgplot.pgtbox('',0,0,'ATS',0,0)

      ppgplot.pgunsa()

   def _plot_labels(self, rangex, rangey, just):
      '''Plot the title, xlabels and ylabels.  This needs the justifiation
         parameter, so that we don't do things with certain Panels.'''

      ppgplot.pgsave()
      ppgplot.pgsch(self.options['fsize'])
      ppgplot.pgscf(self.options['font'])
      ppgplot.pgslw(self.options['width'])
      _set_color(self.options['labelcolor'])

      if self.options['title'] is not None and just not in [4,5,6,7,8,9,13,14]: 
         # First, we see if the title will fit in the viewport.  See how long
         # the string is in world coordinates:
         res = ppgplot.pgqtxt(0,0,0,0.5,self.options['title'])
         xwpc = (res[0][2] - res[0][0])/len(self.options['title'])
         max_length = abs(int((rangex[1]-rangex[0])/xwpc))
         if max_length == 0:  max_length = 10.0
         strs = tw.wrap(self.options['title'], max_length)
         if len(strs) > 3:  strs = strs[0:3]
         strs.reverse()
         # how far do we have to displace vertically to avoid x2range and
         #  x2label?
         i = 1
         num_chars = 1  ;# default displacement when nothing is there.
         if self.options['x2range']:  num_chars = num_chars + 1
         if self.options['x2label']:  num_chars = num_chars + 1
         # factor = 1.0/len(strs)
         for str in strs:
            y = (num_chars + 1 - len(strs)) + i
            ppgplot.pgmtxt("T", y, 0.5, 0.5, str)
            i = i + 1

      if just not in [1,2,3,4,5,6,12,14] and self.options['xlabel']:
         ppgplot.pgmtxt("B", 2.6, 0.5, 0.5, self.options['xlabel'])
      if just not in [2,3,5,6,8,9,11,15] and self.options['ylabel']:   
         ppgplot.pgmtxt("L", 2.6, 0.5, 0.5, self.options['ylabel'])
      if (self.options['x2label'] and just not in [4,5,6,7,8,9,13,14]):
         num_chars = 1
         if self.options['x2range']:  num_chars = num_chars + 1
         ppgplot.pgmtxt("T", num_chars, 0.5, 0.5, self.options['x2label'])
      if (self.options['y2label'] and just not in [1,2,4,5,7,8,10,15]):
         ppgplot.pgmtxt("R", 3.0, 0.5, 0.5, self.options['y2label'])
      # Add ID line if required
      if not just and (self.options['id']==1): ppgplot.pgiden()
      ppgplot.pgunsa()

   # Look through the data and find the range that contains it all.
   def _update_limits(self):
      if not self.lines and not self.points and not self.contours and not self.images and not self.colorpoints and not self.hists and not self.errors:
         self._xmin = 0.0
         self._xmax = 1.0
         self._ymin = 0.0
         self._ymax = 1.0
      else:
         data = []
         hold = []
         # This will deal with custom ordering.
         for item in self.lines + self.points + self.contours + self.images + \
               self.hists + self.colorpoints + self.errors:
            if item.order == -1: 
               data.append(item)
            else:
               hold.append(item)
         for item in hold:
            data.insert(item.order, item)

         self._xmin = None
         self._xmax = None
         self._ymin = None
         self._ymax = None
         for item in data:
            if item.autorange:
               ((xmin,xmax),(ymin,ymax)) = item.range()
               if self._xmin is None:  
                  self._xmin = xmin
               else:
                  self._xmin = min([self._xmin,xmin])
               if self._xmax is None:
                  self._xmax = xmax
               else:
                  self._xmax = max([self._xmax,xmax])
               if self._ymin is None:
                  self._ymin = ymin
               else:
                  self._ymin = min([self._ymin,ymin])
               if self._ymax is None:
                  self._ymax = ymax
               else:
                  self._ymax = max([self._ymax,ymax])

      if self._xmin is None:  self._xmin = 0
      if self._xmax is None:  self._xmax = 1
      if self._ymin is None:  self._ymin = 0
      if self._ymax is None:  self._ymax = 1

   def plot(self, just=0):
      '''Plot the plot object.  This is called if the Plot instance is not
         part of a MPlot or Panel.  Otherwise, MPlot or Panel will be
         responsible for calling it.  The optional just parameter is used
         internally.'''
      self.just = just
      # global _ppgplot_dev_open
      # if not _ppgplot_dev_open:
      # if we're not part of an Mplot or Panel, we open a device.
      if self._panel_num is None and self._id is None:
         self._id = ppgplot.pgopen(self.options['device'])
         #_ppgplot_dev_open = 1
         self._own_dev = 1
         if self.options['aspect']:
            ppgplot.pgpap(0.0, self.options['aspect'])

      self._update_limits()
      if self.options['xrange']:
         rangex = self.options['xrange']
      else:
         if self.options['flipxaxis']:
            rangex = [self._xmax,self._xmin]
         else:
            rangex = [self._xmin,self._xmax]
      if self.options['yrange']:
         rangey = self.options['yrange']
      else:
         if self.options['flipyaxis']:
            rangey = [self._ymax, self._ymin]
         else:
            rangey = [self._ymin, self._ymax]

      #if self.options['flipxaxis']:  (rangex[1], rangex[0]) = rangex
      #if self.options['flipyaxis']:  (rangey[1], rangey[0]) = rangey

      # set the window for all the plotting to be done.
      self._set_viewport(self.just)
      self._set_window(rangex, rangey, self.just, 0)
      (self.xmin,self.xmax) = rangex
      (self.ymin,self.ymax) = rangey

      all_items = []
      hold = []
      # This will deal with custom ordering.
      for item in self.images + self.lines + self.errors + self.points +\
                  self.labels + self.contours + self.hists + self.axes + \
                  self.colorpoints:
         if item.order == -1: 
            all_items.append(item)
         else:
            hold.append(item)
      for item in hold:
         all_items.insert(item.order, item)
      #all_items = self.images + self.lines + self.errors + self.points +\
      #   self.labels + self.contours + self.hists + self.axes + \
      #   self.colorpoints
      for data in all_items:
         data.plot()
      if self.the_legend:  self.the_legend.plot()

      # Now that all the plotting is done, we do the box (so that we know
      #  it's always on top
      self._draw_box(rangex, rangey, self.just)
      self._plot_labels(rangex, rangey, self.just)

   def world2vp(self, x, y):
      '''Converts the (x,y) pair to normalized device coordinates.  The viewport
      and window must be setup (i.e., plot() must have been called).'''
      if self.parent is None:
         u1,u2,v1,v2 = ppgplot.pgqvp(0)
      else:
         u1,u2,v1,v2 = self.parent.layout.pgqvp(self._panel_num)
      x1 = self.xmin;  x2 = self.xmax;  y1 = self.ymin;  y2 = self.ymax
      u = u1 + (x - x1)*(u2 - u1)/(x2 - x1)
      v = v1 + (y - y1)*(v2 - v1)/(y2 - y1)
      return(u,v)

   def vp2world(self, u, v):
      '''Converts the (u,v) pair to world coordinates.  The viewport
      and window must be setup (i.e., plot() must have been called).'''
      if self.parent is None:
         u1,u2,v1,v2 = ppgplot.pgqvp(0)
      else:
         u1,u2,v1,v2 = self.parent.layout.pgqvp(self._panel_num)
      #print u1,u2,v1,v2
      x1 = self.xmin;  x2 = self.xmax;  y1 = self.ymin;  y2 = self.ymax
      x = x1 + (u - u1)*(x2 - x1)/(u2 - u1)
      y = y1 + (v - v1)*(y2 - y1)/(v2 - v1)
      return(x,y)


   def interact(self, bindings=None, mode=7):
      '''Select the plots device, and enter an interactive loop.  Default
      bindings include:  "z,Z":  zoom in, zoom out
                         "A":    (left click) print coordinates
                         "X":    (right click) re-center
                         "r":    redraw
                         'x':    select xrange
                         'y':    select yrange
                         'b':    select x and y range (box)

      make new bindings (or override existing ones) with the bindings keyword.
      The only one you can't override is 'q', to quite the interactive session.
      It is a dictionary, each key representing a key-press.  The associated
      data should be a function accepting the parameters x,y,key,plot  
      where x and y are the coordinates, key is the key pressed and plot is 
      the instance of the plot.'''

      binds =    {'A': printcoords,
                  'X': pan,
                  'r': redraw,
                  'Z': zoom,
                  'z': zoom,
                  'x': xrange_select,
                  'y': yrange_select,
                  'b': box_range_select}
      if bindings:
         for key in bindings:
            binds[key] = bindings[key]

      x = y = 0
      key = ""
      while key != 'q':
         self.select()
         self.replot()
         x,y,key = ppgplot.pgband(mode,0,x,y)
         if key in binds.keys():
            ret = apply(binds[key], (x,y,key,self))
            if ret is not None:
               return(ret)




   def replot(self):
      '''Select the plot's device, erase and redraw the plot.  Use this in
         the interactive mode to make changes and replot.'''
      self.select()
      ppgplot.pgeras()
      self.plot()
   
   # Select this plot's device (and maybe panel) so that ppgplot commands go
   #  to its viewport
   def select(self):
      '''This will make the instance's device the selected device (and make
         it's panel the current pane if it is part of an Mplot or Panel).  All
         ppgplot raw commands will then be directed to this plot.'''
      if not self._id:
         print "Sorry, I don't seem to be attached to any device.  Try plot()"
         return -1
      ppgplot.pgslct(self._id)
      if self._panel_num is not None:
         # remember, the coordinates are numbered from one (sheesh)
         # ppgplot.pgpanl(self._panel_num_i+1, self._panel_num_j+1)
         self.parent.layout.select(self._panel_num)

   # Close the plot (close the device if it is owned by the plot)
   def close(self):
      '''Close the plot device.  Only effective if the Plot instance is not
         part of a MPlot or Panel instance.'''
      global _ppgplot_dev_open
      if self._own_dev:
         ppgplot.pgslct(self._id)
         ppgplot.pgclos()
         #_ppgplot_dev_open = 0
         self._own_dev = 0
         self._id = None

   def __setattr__(self,name,value):
      if name in self.options.keys():
         self.options[name] = value
      else:
         self.__dict__[name] = value


class MPlot:
   '''Class used to create a multi-panel plot with separate bounding boxes.
      This is in the traditional
      sense, where a page is split into (n x m) panels and each panel has
      a full plot:  box, axes, tick lables, axis labels, titles, etc.
      Required parameters are the number of columns and rows.  The user then
      adds Plot instances which will be ploted in the sub-panels
      in order.

      Example:
      
      >>> M = MPlot(2,2,device="test.ps/PS")    # create a 2 x 2  multi-plot
      >>> M.add(plot1)     # plot1, plot2, plot3 and plot4 are Plot 
      >>> M.add(plot2)     #  instances
      >>> M.add(plot3)
      >>> M.add(plot4)
      >>> M.plot()
      >>> M.close()
   '''

   def __init__(self, nx=1, ny=1, device=_default_device, aspect=None,
         llcs=None, urcs=None):
      '''Initialization routine for the MPlot class.
         @param nx: number of columns
         @type nx:  int
         @param ny: number of rows
         @type ny:  int
         @param device:  Pgplot device to use (default:  1st available device)
         @type device: string
         @param aspect:  The aspect ratio (height/width) (default:  device
                         dependent).
         @type aspect: float
         @param llcs:  The lower-left corners of the sub-plots.
         @type llcs:  list
         @param urcs:  the upper-right corners of the sub-plots.
         @type urcs:  list
      '''
      self._own_dev = 0
      self.nx = 1           # Number of x panels
      self.ny = 1           # Number of y panels
      self._id = None
      self.nx = nx
      self.ny = ny
      self.device = device
      self.aspect = aspect
      self.plots = []       # list holder for the actual plots
      self.layout = _Layout()
      # if the user specifies nx and ny, use equal-size panels, otherwise,
      # use the llcs and urcs provided
      if llcs is None and urcs is None:
         xsize = 1.0/nx
         ysize = 1.0/ny
         for j in range(ny):
            y0 = (ny - 1 - j)*ysize;  y1 = (ny - j)*ysize
            for i in range(nx):
               x0 = i*xsize;  x1 = (i+1)*xsize
               self.layout.add_panel((x0,y0), (x1,y1))
      else:
         for i in range(len(llcs)):
            self.layout.add_panel(llcs[i], urcs[i])


   def __del__(self):
      # ensure the device is closed
      self.close()

   def add(self, plot):
      '''Add a Plot instance to the MPlot.  The instances are plotted
         in the order they are added, filling each row and then switching to
         the next.
      '''
      #plot._panel_num_i = len(self.plots) % self.nx
      #plot._panel_num_j = len(self.plots) / self.nx
      plot._panel_num = len(self.plots)
      plot.parent = self
      self.plots.append(plot)

   def plot(self):
      '''Plot the Mplot.  Each sub-panel is plotted by calling that instance's
         plot() method after calling pgpage() to advance to the next panel.
      '''
      #global _ppgplot_dev_open
      if self._id is None:
         self._id = ppgplot.pgopen(self.device)
         #_ppgplot_dev_open = 1
         self._own_dev = 1
         if self.aspect:
            ppgplot.pgpap(0.0, self.aspect)

      self.layout.current = -1
      # Do this with Layout class now
      #ppgplot.pgsubp(self.nx,self.ny)

      for plot in self.plots:
         self.layout.pgpage()
         plot._id = self._id
         plot.plot()

   def replot(self):
      ppgplot.pgeras()
      for plot in self.plots:
         plot.select()
         plot.plot()

   def interact(self, bindings=None, mode=7):
      '''This is similar to Plot's interact, but is more of an overseer:  it 
      figures
      out which panel the event is supposed to happen in, then calls that Plot's
      binding do the the work.
      '''

      binds =    {'A': printcoords,
                  'X': pan,
                  'r': redraw,
                  'Z': zoom,
                  'z': zoom,
                  'x': xrange_select,
                  'y': yrange_select,
                  'b': box_range_select}
      if bindings:
         for key in bindings:
            binds[key] = bindings[key]

      x = y = 0
      key = ""
      p0 = self.plots[self.layout.current]
      while key != 'q':
         self.replot()
         x,y,key = ppgplot.pgband(mode,0,x,y)
         u,v = p0.world2vp(x,y)
         i = self.layout.which_vp(u,v)
         if i > len(self.plots) - 1:  continue
         if i != -1 and key in binds.keys():
            cp = self.plots[i]
            cp.replot()
            x,y = cp.vp2world(u,v)
            ret = apply(binds[key], (x,y,key,cp))
            self.replot()
            if ret is not None: return(ret)

   # Select the the device and optionally a panel to make current
   def select(self, i=0):
      '''Make the instance's device the current device and select the given
         panel.
         @param i: which panel to select (default:  0)
         @type i: int
      '''
      if not self._id:
         print "Sorry, I don't seem to be attached to any device.  Try plot()"
         return -1
      ppgplot.pgslct(self._id)
      #ppgplot.pgpanl(i+1, j+1)
      self.layout.select(i)
      

   # Close the device if we own it.
   def close(self):
      '''Close the pgplot device corresponding to the MPlot.'''
      global _ppgplot_dev_open
      if self._own_dev and self._id:
         ppgplot.pgslct(self._id)
         ppgplot.pgclos()
         #_ppgplot_dev_open = 0
         self.own_dev = 0
         self._id = None

class Panel:
   '''Class used to create a multi-panel plot with a single bounding box.
      Unlike MPlot, these panels
      are made to look like one plot with contiguous sub-plots.  For example,
      if the plot were split into 3 x 3 panels, the central panel would have
      no labels outside it's box, which would abutt the sides of the adjoining
      panels.  Other than this property, this class works the same as MPlot.
   '''

   def __init__(self, nx=1, ny=1, device=_default_device, aspect=None,
         llcs=None, urcs=None, justs=None):
      '''Initialization routine for the MPlot class.
         @param nx: number of columns
         @type nx:  int
         @param ny: number of rows
         @type ny:  int
         @param device:  Pgplot device to use (default:  1st available device)
         @type device: string
         @param aspect:  The aspect ratio (height/width) (default:  device
                         dependent).
         @type aspect: float
         @param urcs:  The upper-right corners of the Panels
         @type urcs:  list of 2-tuples
         @param llcs:  The lower-left corners of the Panels
         @type llcs:  list of 2-tuples
      '''
      self._own_dev = 0
      self.nx = 1           # Number of x panels
      self.ny = 1           # Number of y panels
      self._id = None
      self.nx = nx
      self.ny = ny
      self.device = device
      self.aspect = aspect
      self.plots = []

      self.layout = _Layout()
      # For now, we just put in nx X ny panels of equal size.  We'll add borders later,
      #   when we've figured out if there is need.
      if llcs is None and urcs is None:
         self.justs = []
         xsize = 1.0/nx
         ysize = 1.0/ny
         n = 0
         for j in range(ny):
            y0 = (ny - 1 - j)*ysize;  y1 = (ny - j)*ysize
            for i in range(nx):
               x0 = i*xsize;  x1 = (i+1)*xsize
               self.layout.add_panel((x0,y0), (x1,y1))
               self.justs.append(self.get_just(n))
               n += 1
      else:
         if justs is None:
            justs = [0]*len(urcs)
         if len(justs) != len(llcs) or len(urcs) != len(llcs):
            raise ValueError, "llcs, urcs, and justs must be same length"
         for i in range(len(llcs)):
            self.layout.add_panel(llcs[i], urcs[i])
         self.justs = justs

   def __del__(self):
      self.close()
      
   def add(self, plot):
      '''Add a Plot instance to the MPlot.  The instances are plotted
         in the order they are added, filling each row and then switching to
         the next.
      '''
      #plot._panel_num_i = len(self.plots) % self.nx
      #plot._panel_num_j = len(self.plots) / self.nx
      plot._panel_num = len(self.plots)
      plot.parent = self
      self.plots.append(plot)

   def get_just(self, i):
      '''Get the correct justification for panel i.'''
      pan_per_page = self.nx*self.ny
      row = (i%pan_per_page)/self.nx   # row of the plot
      col = (i%pan_per_page)%self.nx   # column of the plot
      
      if self.nx == 1 and self.ny ==1:  #  The first and last panels:
         just = 0
      elif self.nx == 1:
         if row == 0: just = 12
         elif row == self.ny - 1:  just=13
         else:  just=14
      elif self.ny == 1:
         if col == 0:  just = 10
         elif col == self.nx - 1:  just=11
         else:  just=15
      else:
         if row == 0:
            if col == 0: just = 1
            elif col == self.nx-1: just = 3
            else: just = 2
         elif row == self.ny - 1:
            if col == 0:  just = 7
            elif col == self.nx-1: just = 9
            else: just = 8
         else:
            if col == 0: just = 4
            elif col == self.nx-1:  just = 6
            else: just=5
      return(just)

   def plot(self):
      '''Plot the Panel.  Each sub-panel is plotted by calling that instance's
         plot() method after calling pgpage() to advance to the next panel.
      '''
      #global _ppgplot_dev_open
      if self._id is None:
         self._id = ppgplot.pgopen(self.device)
         #_ppgplot_dev_open = 1
         self._own_dev = 1
         if self.aspect:
            ppgplot.pgpap(0.0, self.aspect)

      self.layout.current = -1     # Start off at first page
      # Now, go through the plots and count up how much border space we're 
      # likely to need
      xb1 = xb2 = by1 = by2 = 0
      for plot in self.plots:
         vx1,vx2,vy1,vy2 = plot._set_viewport(just=0, compute_only=1)
         if vx1 > xb1:  xb1 = vx1
         if (1-vx2) > xb2:  xb2 = (1 - vx2)
         if vy1 > by1:  by1 = vy1
         if (1-vy2) > by2:  by2 = (1 - vy2)
      self.layout.add_border(xb1,xb2,by1,by2)

      for i in range(len(self.plots)):
         plot = self.plots[i]
         #ppgplot.pgpage()
         self.layout.pgpage()

         plot._id = self._id
         just = self.justs[i]
         #just = self.get_just(i)
         plot.plot(just=just)

   def replot(self):
      ppgplot.pgeras()
      for i in range(len(self.plots)):
         #just = self.get_just(i)
         just = self.justs[i]
         self.plots[i].select()
         self.plots[i].plot(just=just)


   def interact(self, bindings=None, mode=7):
      '''This is similar to Plot's interact, but is more of an overseer:  it 
      figures
      out which panel the event is supposed to happen in, then calls that Plot's
      binding do the the work.
      '''

      binds =    {'A': printcoords,
                  'X': pan,
                  'r': redraw,
                  'Z': zoom,
                  'z': zoom,
                  'x': xrange_select,
                  'y': yrange_select,
                  'b': box_range_select}
      if bindings:
         for key in bindings:
            binds[key] = bindings[key]

      x = y = 0
      key = ""
      p0 = self.plots[self.layout.current]
      while key != 'q':
         self.replot()
         x,y,key = ppgplot.pgband(mode,0,x,y)
         u,v = p0.world2vp(x,y)
         i = self.layout.which_vp(u,v)
         if i > len(self.plots) - 1:  continue
         if i != -1 and key in binds.keys():
            cp = self.plots[i]
            cp.replot()
            x,y = cp.vp2world(u,v)
            ret = apply(binds[key], (x,y,key,cp))
            self.replot()
            if ret is not None: return(ret)
            
   # Select the the device and optionally a panel to make current
   def select(self, i=0):
      '''Make the instance's device the current device and select the given
         panel.
         @param i: which panel to select (default:  0)
         @type i: int
      '''
      if not self._id:
         print "Sorry, I don't seem to be attached to any device.  Try plot()"
         return -1
      ppgplot.pgslct(self._id)
      self.layout.select(i)
      #ppgplot.pgpanl(i+1, j+1)
      

   # Close the device if we own it.
   def close(self):
      '''Close the pgplot device corresponding to the MPlot.'''
      #global _ppgplot_dev_open
      if self._own_dev:
         ppgplot.pgslct(self._id)
         ppgplot.pgclos()
         self.own_dev = 0
         self._id = None


# Data should be a Numeric Array
def _scalerange(data):
    """
    Adjust the range to be plotted so that it fits nicely on the page.
    Return a list with adjusted minimum and maximum values from 'data'.
    """
    #min = Numeric.minimum.reduce(data)
    #max = Numeric.maximum.reduce(data)
    min = nummin(Numeric.ravel(data))
    max = nummax(Numeric.ravel(data))
    extra = 0.1 * (max - min)
    return [min - extra, max + extra]

hex_pat = re.compile('([A-Z0-9]{2,2})([A-Z0-9]{2,2})([A-Z0-9]{2,2})')
def _set_color(color):
   if (type(color) is types.StringType):
      if color in _color_table:
         ppgplot.pgsci(_color_table[color])
      else:
         result = hex_pat.match(color)
         if result is not None:
            R,G,B = result.groups()
            R = string.lower(R); G = string.lower(G); B = string.lower(B)
            R,G,B = map(lambda x: float(int(x,16))/255, [R,G,B])
            ppgplot.pgscr(16, R, G, B)
            ppgplot.pgsci(16)
         else:
            raise AttributeError, "Color %s not defined" % (color)
   else:
      ppgplot.pgsci(color)

class _Palette:
    # Set the color palette
    def __init__(self):
       self.rgbs = colormaps.colormaps.keys()
       self.transfers = colormaps.transfers.keys()
    def setpalette(self, palette, transfer='ramp'):
        """
        Set the color palette for imag-style routines
        """
        if palette not in self.rgbs:
           raise ValueError, "specified palettte '%s' not defined." % (palette)
        if transfer not in self.transfers:
           raise ValueError, "specified transfer function '%s' not defined" \
                 % (transfer)

        self.l = colormaps.transfers[transfer]
        self.r = colormaps.colormaps[palette][:,0]
        self.g = colormaps.colormaps[palette][:,1]
        self.b = colormaps.colormaps[palette][:,2]

#        if (palette == 'rainbow'):
#            self.l = Numeric.array([0.0, 0.015, 0.225, 0.4, 0.59,
#                                    0.6, 0.775, 0.955, 0.965, 1.0])
#            self.r = Numeric.array([1.0, 1.0, 1.0, 0.0, 0.0,
#                                    0.0, 0.0, 0.947, 1.0, 1.0])
#            self.g = Numeric.array([0.0, 0.0, 1.0, 1.0, 1.0,
#                                    0.946, 0.0, 0.8, 0.844, 1.0])
#            self.b = Numeric.array([0.0, 0.0, 0.0, 0.0, 0.95,
#                                    1.0, 1.0, 1.0, 1.0, 1.0])
#        if (palette == 'antirainbow'):
#            self.l = Numeric.array([0.0, 0.035, 0.045, 0.225, 0.4, \
#                                    0.41, 0.6, 0.775, 0.985, 1.0])
#            self.r = Numeric.array([1.0, 1.0, 0.947, 0.0, 0.0,
#                                    0.0, 0.0, 1.0, 1.0, 1.0])
#            self.g = Numeric.array([1.0, 0.844, 0.8, 0.0, 0.946,
#                                    1.0, 1.0, 1.0, 0.0, 0.0])
#            self.b = Numeric.array([1.0, 1.0, 1.0, 1.0, 1.0,
#                                    0.95, 0.0, 0.0, 0.0, 0.0])
#        elif (palette == 'astro'):
#            self.l = Numeric.array([0.0, 0.167, 0.333, 0.5,
#                                    0.667, 0.833, 1.0])
#            self.r = Numeric.array([0.0, 0.0, 0.0, 0.0, 1.0, 1.0, 1.0])
#            self.g = Numeric.array([0.0, 0.0, 1.0, 1.0, 1.0, 0.0, 1.0])
#            self.b = Numeric.array([0.0, 1.0, 1.0, 0.0, 0.0, 0.0, 1.0])
#        elif (palette == 'hue'):
#            self.l = Numeric.array([0.0, 0.167, 0.333, 0.5,
#                                    0.667, 0.833, 1.0])
#            self.r = Numeric.array([1.0, 1.0, 0.0, 0.0, 0.0, 1.0, 1.0])
#            self.g = Numeric.array([0.0, 1.0, 1.0, 1.0, 0.0, 0.0, 0.0])
#            self.b = Numeric.array([0.0, 0.0, 0.0, 1.0, 1.0, 1.0, 0.0])
#        elif (palette == 'heat'):
#            self.l = Numeric.array([0.0, 0.48, 0.7, 0.75, 1.0])
#            self.r = Numeric.array([0.0, 1.0, 1.0, 1.0, 1.0])
#            self.g = Numeric.array([0.0, 0.0, 0.423, 0.519, 1.0])
#            self.b = Numeric.array([0.0, 0.0, 0.0, 0.0, 1.0])
#        elif (palette == 'gamma'):
#            self.l = Numeric.array([0.0, 0.33, 0.66, 1.0])
#            self.r = Numeric.array([0.3, 1.0, 0.0, 0.0])
#            self.g = Numeric.array([0.0, 0.3, 1.0, 0.0])
#            self.b = Numeric.array([0.0, 0.0, 0.3, 1.0])
#        elif (palette == 'antigray' or palette == 'antigrey'):
#            self.l = Numeric.array([0.0, 1.0])
#            self.r = Numeric.array([1.0, 0.0])
#            self.g = Numeric.array([1.0, 0.0])
#            self.b = Numeric.array([1.0, 0.0])
#        else:
#            self.l = Numeric.array([0.0, 1.0])
#            self.r = Numeric.array([0.0, 1.0])
#            self.g = Numeric.array([0.0, 1.0])
#            self.b = Numeric.array([0.0, 1.0])
#
_pgpalette = _Palette()

def list_palettes():
   '''Print the currently installed palettes to the terminal.'''
   print 'Available palettes:'
   print colormaps.colormaps.keys()
   print 'Available transfers:'
   print colormaps.transfers.keys()

def _spline (x, c0, c1, c2, c3):
    n = len(x) - 1

    d = Numeric.zeros((n+1), Numeric.Float64)
    diag = Numeric.zeros((n+1), Numeric.Float64)
    d[1:] = x[1:] - x[:-1]
    d[0] = 0.0
    diag[1:] = (c0[1:] - c0[:-1])/d[1:]
    diag[0] = 1.0


    c1[1:-1] = 3.0*(d[1:-1]*diag[2:] + d[2:]*diag[1:-1])
    diag[1:-1] = 2.0*(d[1:-1] + d[2:])

    for m in range(1,n):
       g = -d[m+1]/diag[m-1]
       diag[m] = diag[m] + g*d[m-1]
       c1[m] = c1[m] + g*c1[m-1]

    for m in range(n-1,0,-1):
        c1[m] = ( c1[m] - d[m]*c1[m+1] )/diag[m]

    # dx = x[1:] - x[:-1]
    # d1 = (c0[1:] - c0[:-1])/dx
    # d3 = c1[:-1] + c1[1:] - 2.0*d1
    # c2[:-1] = (d1 - c1[:-1] - d3)/dx
    # c3[:-1] = d3/(dx*dx)

    for m in range (0,n):
        dx = x[m+1] - x[m]
        d1 = ( c0[m+1] - c0[m] )/dx
        d3 = c1[m] + c1[m+1] - 2.0*d1
        c2[m] = ( d1 - c1[m] - d3 )/dx
        c3[m] = d3/(dx*dx)
    dx = x[:-1]
    c0[:-1] = c0[:-1] - dx*(c1[:-1] - dx*(c2[:-1] - dx*c3[:-1]))
    c1[:-1] = c1[:-1] - dx*(2.0*c2[:-1] - 3.0*dx*c3[:-1])
    c2[:-1] = c2[:-1] - 3.0*dx*c3[:-1]
    #for m in range (0,n):
    #    dx = x[m]
    #    c0[m] = c0[m] - dx*(c1[m] - dx*(c2[m] - dx*c3[m]))
    #    c1[m] = c1[m] - dx*(2.0*c2[m] - 3.0*dx*c3[m])
    #    c2[m] = c2[m] - 3.0*dx*c3[m]

#/*
#evaluate the spline of the n-interval (n+1 points)  data vector, x, using
#the previously derived spline coefficient vectors c0[], ..., c3[].
#*/
def _evspline (xp, x, c0, c1, c2, c3):
    n = len(x)
    if xp < x[0]: 
       n = 0 
    else: 
        n = n-1 #        /* use highest interval for all higher values ... */
        while ( x[n] > xp ):  n = n-1  #      /* until we get below xp */
        
    return c0[n] + xp*(c1[n] + xp*(c2[n] + xp*c3[n])) 
#/*
#PGSPLINE:        (CCD)
#Purpose:
#    Take "rough" input points, and return values for requested x-values,
#    using cubic splines, with given end-slopes.
#Input:
#    x-y vector of points, np such points,
#    start slope = slope0, end slope = slope1.
#    vector x of n values where values of y are required.
#Output:
#    vector y, the values at these n points
#
#                                NB: input xp, yp are not altered.
#*/
def pgspline (xp, yp, x, slope0=None, slope1=None):
   '''Compute the cubic spline for a set of data
   @param xp:  An array of x-coordinates for the control points.
   @type xp:  float array
   @param yp:  An array of y-coordinates for the control points.
   @type yp:  float array
   @param x:  An array of x-coordinates where you want interpolations to be
              done
   @type x:  float array
   @param slope0:  Optionally specify the slope at the beginning of the interval
   @type slope0:  float
   @param slope1:  Optionally specify the slope at the end of the interval
   @type slope1:  float
   '''
   xp = Numeric.asarray(xp, Numeric.Float64)
   yp = Numeric.asarray(yp, Numeric.Float64)
   x = Numeric.asarray(x, Numeric.Float64)
   y = x*1.0
   np = len(xp)
   n = len(x)
   c0 = Numeric.zeros((np), Numeric.Float64)
   c1 = Numeric.zeros((np), Numeric.Float64)
   c2 = Numeric.zeros((np), Numeric.Float64)
   c3 = Numeric.zeros((np), Numeric.Float64)

   if slope0 is None:
      slope0 = (yp[1] - yp[0])/(xp[1] - xp[0])
   if slope1 is None:
      slope1 = (yp[-1] - yp[-2])/(xp[-1] - xp[-2])

   invert = 0;
   if( xp[1] < xp[0] ): invert = 1;
   if( invert ):
      xp = -xp;
      x = -x;
   c0[0] = yp[0]
   for i in range(1,np):
      c0[i] = yp[i]
      if( xp[i] < xp[i-1] ):
         sys.stderr.write("PGSPLINE: non-monotonic input encountered\n")
         if( invert ):
            xp = -xp;
            x = -x;
         return(0);
   c1[0] = slope0
   c1[np-1] = slope1
   _spline(xp,c0,c1,c2,c3)
   for i in range(0,n): y[i] = _evspline(x[i],xp,c0,c1,c2,c3)
   if( invert ): 
      xp = -xp
      x = -x
   return(y)


def save(instance, file):
   '''Save a plot instance in file.  This uses python's pickle module.'''
   f = open(file, 'w')
   pickle.dump(instance, f)
   f.close()

def load(file):
   '''Load a previously saved plot instance from file.'''
   f = open(file, 'r')
   instance = pickle.load(f)
   f.close()
   return(instance)
   

def columns(filename, skip=[], delim=None):
   '''Read in the specified file and try to extract columns.  So far, it will
      recognise space-, comma-, tab-, or semi-colon-delimited columns.  
      Columns are numbered from 0.
      @param filename:  the name of the file to read
      @type filename:  string
      @param skip:  List of lines to skip (besides lines commented with '#')
      @type skip:  list of integers
      @param delim:  Specify the delimeter explicitly
      @type delim:  string
      @return:  a 2-D array:  one row for each column in the file
   '''
   f = open(filename,'r')
   if not f:
      raise IOError, 'Could not find file %s' % (filename)

   lines = f.readlines()
   data = []
   i = 0
   for line in lines:
      line = string.strip(line)
      if line == "":  continue
      if line[0] != "#" and i not in skip:  data.append(line)
      i = i + 1

   if delim is None:
      line = data[0]
      if len(string.split(line)) > 1:
         sep = None
      elif len(string.split(line,',')) > 1:
         sep = ","
      elif len(string.split(line,';')) > 1:
         sep = ";"
      else:
         sep = None
   else:
      sep = delim

   data = map(lambda str,s=sep: string.split(str, s), data)

   #check for consistency:
   M = len(data)
   N = len(data[0])
   for line in data:
      if len(line) != N:
         raise IndexError, "data has missing cells"

   # now do a transpose
   datat = []
   for i in range(N):
      datat.append([])
      for j in range(M):
         datat[-1].append(data[j][i])

   # Now try to convert to floats and numarrays
   for i in range(len(datat)):
      try:
         datat[i] = map(float, datat[i])
         datat[i] = Numeric.array(datat[i])
      except:
         # silently ignore non-numeric types.
         datat[i] = map(None, datat[i])

   return datat

_plot = None
_data = None
def plot(filename, u=[0,1], w=['l','p'], skip=[], delim=None):
   '''Convenience function to generate out a plot instance using columnar data
   from filename, using columns specified in using (a list of 2-tuples, one for
   each data line).'''
   global _plot, _data
   _data = columns(filename, skip, delim) 
   if type(u[0]) is not types.ListType:
     u = [u]

   _plot = Plot()

   i = 0
   for x,y in u:
     if 'l' in w:
        _plot.line(_data[x],_data[y], color=i+1)
     if 'p' in w:
        _plot.point(_data[x],_data[y], color=i+1, symbol=i+2)

   _plot.plot()
   return(_plot)

def replot():
   global _plot,_data
   if _plot is not None: _plot.replot()

def set(option,value):
   global _plot,_data
   if _plot is not None:
      _plot.__dict__[option] = value

def close():
   global _plot
   if _plot is not None:
      _plot.close()

